import dash_bootstrap_components as dbc

theme_stylesheet = dbc.themes.DARKLY

# Based on https://bootswatch.com/darkly/
colors = [
    "#375a7f",
    "#3a3a3a",
    "#00bc8c",
    "#f39c12",
    "#e74c3c",
    "#3498db",
    "#ad05bd",
]
plot_font_color = "#fff"
plot_legend_font_color = "#fff"
plot_paper_bgcolor = "rgb(48, 48, 48)"
plot_plot_bgcolor = "rgb(173, 181, 189)"

dark_plot_layout = dict(
    font_color=plot_font_color,
    legend_font_color=plot_legend_font_color,
    paper_bgcolor=plot_paper_bgcolor,
    plot_bgcolor=plot_plot_bgcolor,
    margin=dict(b=40, t=40, l=40, r=40),
)

table_style_header = {
    "backgroundColor": "rgb(30, 30, 30)",
    "color": "white",
}

table_style_filter = {
    "backgroundColor": "rgb(40, 40, 40)",
    "color": "white",
}

table_style_data = {
    "backgroundColor": "rgb(50, 50, 50)",
    "color": "white",
}

table_style_cell = {
    "textAlign": "left",
    "textOverflow": "ellipsis",
    "maxWidth": 0,
}

table_active_cell_conditional = {
    "if": {
        "state": "active",
    },
    "color": "red !important",
    "backgroundColor": "rgb(70, 70, 70)",
    "textAlign": "left",
    "borderColor": "#039 !important",
}



import pandas as pd

from entropylab.pipeline.api.param_store import ParamStore


def param_store_to_df(ps: ParamStore):
    ps_dict = ps.to_dict()
    tag_list = []
    for key in ps.keys():
        tags = ps.list_tags_for_key(key)
        tag_list.append(",".join(tags))
    params_df = pd.DataFrame(
        {"key": ps_dict.keys(), "value": ps_dict.values(), "tag": tag_list}
    )
    return params_df


def param_store_to_commits_df(ps: ParamStore):
    commit_list = ps.list_commits()
    ids = []
    labels = []
    times = []
    for commit in commit_list:
        ids.append(commit.id)
        labels.append(commit.label)
        times.append(pd.Timestamp(commit.timestamp, unit="ns").strftime("%Y-%m-%d %X"))
    return pd.DataFrame(
        {"commit_id": ids, "commit_label": labels, "commit_time": times}
    )


def data_diff(params: ParamStore, data, data_prev):
    if len(data) > len(data_prev):
        return data_prev[-1]
    if len(data) < len(data_prev):
        del_key = set([x["key"] for x in data_prev]).difference(
            set([x["key"] for x in data])
        )
        del params[del_key.pop()]
    for idx, _row in enumerate(data):
        if (len(data_prev) == len(data)) and not data_prev[idx] == data[idx]:
            if not data_prev[idx]["key"] == data[idx]["key"]:
                params.rename_key(data_prev[idx]["key"], data[idx]["key"])
            params[data[idx]["key"]] = data[idx]["value"]
            if not data_prev[idx]["tag"] == data[idx]["tag"]:
                params.remove_tag(data_prev[idx]["tag"], data[idx]["key"])
                params.add_tag(data[idx]["tag"], data[idx]["key"])



import os
from shutil import copytree, copy

print("\n = = = = = = entropylab example project = = = = =")
print("Writting simple example project in the current directory...")

origin = os.path.join(
    os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "simple"
)
destination = os.getcwd()

file_list = [
    "cheerful_node.py",
    "grumpy_administrator.py",
    "workflow.py",
    "parameters.json",
]

copytree(
    os.path.join(origin, "entropynodes"),
    os.path.join(destination, "entropynodes"),
    dirs_exist_ok=True,
)

for f in file_list:
    copy(os.path.join(origin, f), destination)

print("\t1. First execute\tpython cheerful_node.py\t\tto test single node")
print("\t2. First execute\tpython grumpy_administrator.py\tto test single node")
print("\t3. Then execute\t\tpython workflow.py\t\tto generate list of parameters")
print(
    "\t4. Open and change parameters.json (Null is not acceptable value for "
    "final resolved parameters) "
)
print("\t5. Execute workflow with\tpython -m entropylab.flame.execute ")
print("\t  (to interrupt and terminate execution press Ctrl+C)")
print("\n")



import dash_bootstrap_components as dbc
import shutil
import os
from dash import html

from entropylab.pipeline.results_backend.sqlalchemy.project import (
    project_name,
    project_path,
)


def top_bar(path: str):
    hdd = shutil.disk_usage(path)
    top_bar_text = f"{project_path(path)}"
    top_bar_text += f"| Project size:{os.path.getsize(path) / 2 ** 20:.4f} MB"
    top_bar_text += f" | Free space:{hdd.free // 2 ** 30} GB"
    return dbc.Row(
        dbc.Navbar(
            [
                dbc.Col(
                    dbc.NavbarBrand(
                        html.Img(
                            src="/assets/images/entropy_logo_dark.svg",
                            width=150,
                            id="entropy-logo",
                        ),
                        href="#",
                    ),
                    width="2",
                    id="logo-col",
                ),
                dbc.Col(
                    [
                        html.Div(
                            f"Project: {project_name(path)}",
                            id="project-name",
                        ),
                        html.Div(
                            top_bar_text,
                            id="project-name",
                            style={"fontSize": "11px"},
                        ),
                    ],
                    width="4",
                ),
                dbc.Col(
                    dbc.Row(
                        [
                            dbc.Col(
                                dbc.NavItem(
                                    dbc.NavLink(
                                        "🔬 Experiment Results",
                                        href="/",
                                        active=True,
                                    )
                                )
                            ),
                            dbc.Col(
                                dbc.NavItem(dbc.NavLink("🧮 Params", href="/params"))
                            ),
                        ]
                    ),
                    width="6",
                ),
            ],
            color="primary",
        ),
    )



import entropynodes.library as expNodes
from entropylab.flame.workflow import Workflow

wf = Workflow("Workflow name", description="What this realises in more details")

# now start workflow description

# ...


# end of workflow description

wf.register()



"""create_figures_table

Revision ID: f1ada2484fe2
Revises: 04ae19b32c08
Create Date: 2022-03-17 15:57:28.225697+00:00

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
from sqlalchemy.engine import Inspector

revision = "f1ada2484fe2"
down_revision = "04ae19b32c08"
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)
    tables = inspector.get_table_names()
    # ### commands auto generated by Alembic - please adjust! ###
    if "Figures" not in tables:
        op.create_table(
            "Figures",
            sa.Column("id", sa.Integer(), nullable=False),
            sa.Column("experiment_id", sa.Integer(), nullable=False),
            sa.Column("figure", sa.String(), nullable=False),
            sa.Column("time", sa.DATETIME(), nullable=False),
            sa.ForeignKeyConstraint(
                ["experiment_id"], ["Experiments.id"], ondelete="CASCADE"
            ),
            sa.PrimaryKeyConstraint("id"),
        )


def downgrade():
    op.drop_table("Figures")



from entropylab.flame.inputs import Inputs
from entropylab.flame.workflow import Workflow

__all__ = ["FinalDataAnalysis"]


class FinalDataAnalysis(object):
    def __init__(self, workflow_node_unique_name, data=None):
        """Final fitting and plotting

        :param data: (JSON - STREAM) experiment results
        """
        self._command = "python3"
        self._bin = "plot.py"
        self._name = workflow_node_unique_name
        self._icon = "bootstrap/file-bar-graph.svg"
        self._inputs = _Inputs(
            data=data,
        )
        self._outputs = _Outputs(self._name)
        self._host = {}
        Workflow._register_node(self)  # register the node in the workflow context

    def host(self, **kwargs):
        """Sets additional options for execution on the host."""
        for key, value in kwargs.items():
            self._host[key] = value
        return self

    @property
    def i(self):
        """Node inputs"""
        return self._inputs

    @property
    def o(self):
        """Node outputs"""
        return self._outputs


class _Inputs(object):
    def __init__(self, data=None):
        self._inputs = Inputs()

        self._inputs.state("data", description="experiment results", units="JSON")
        self._inputs.set(data=data)

    @property
    def data(self):
        """Input: experiment results (JSON)"""
        return self._inputs.get("data")

    @data.setter
    def data(self, value):
        """Input: experiment results (JSON)"""
        self._inputs.set(data=value)


class _Outputs(object):
    def __init__(self, name):
        self._name = name
        self._outputs = [
            "estimated_parameter",
            "final_plot",
        ]

    @property
    def estimated_parameter(self):
        """Output: final result
        :return: (MHz)
        """
        return "#" + self._name + "/estimated_parameter"

    @property
    def final_plot(self):
        """Output: to-do
        :return: (png)
        """
        return "#" + self._name + "/final_plot"



import entropylab.flame.nodeio as nodeio
import time

# ==================== DEFINE NODE ====================

nodeio.context(
    name="TaskList",
    description="Does scan of parameters and triggers circuit execution",
    icon="bootstrap/list-task.svg",
)

input = nodeio.Inputs()
input.state(
    "calibration",
    units="status",
    description="node will execute only if it receives positive calibration status",
)
input.state("scan_points", units="MHz", description="an array for setpoint execution")

input.stream(
    "laser_setpoint_locked",
    units="bool",
    description="Triger when laser setpoint is set",
)
input.stream(
    "circuit_done", units="bool", description="Triger when circuit execution is done"
)

output = nodeio.Outputs()
output.define(
    "circuit_specification",
    units="JSON",
    description="Triggers circuit execution by providing information for run",
    retention=0,
)
output.define("scan_setpoint", units="MHz", description="Laser setpoint")
output.define(
    "status", units="json", description="window into what node is doing", retention=1
)

nodeio.register()  # enables this to be used as part of workflow

# ==================== DRY-RUN DATA ====================

input.set(scan_points=[1, 2, 3], calibration="calibrated")
input.set(laser_setpoint_locked=True, circuit_done=True)
input.set(laser_setpoint_locked=True, circuit_done=True)
input.set(laser_setpoint_locked=True, circuit_done=True)

# =============== RUN NODE STATE MACHINE ===============

sequence = input.get("scan_points")
index = 0

while nodeio.status.active:
    status = input.get("calibration")
    if status == "calibrated":
        if index < len(sequence):
            output.set(status={"scan point": sequence[index]})
            print("Set point : ", sequence[index])
            output.set(scan_setpoint=sequence[index])
            setpointReached = input.get("laser_setpoint_locked")
            output.set(circuit_specification=sequence[index])
            circuit_done = input.get("circuit_done")
            time.sleep(2)
        else:
            print("terminating experiment...")
            time.sleep(2)
            nodeio.terminate_workflow()
        index += 1
    else:
        pass


# ================= =================== =================



import os
from shutil import copytree, copy

print("\n = = = = = = entropylab example project = = = = =")
print("Writting simple example project in the current directory...")

origin = os.path.join(
    os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "complex"
)
destination = os.getcwd()

file_list = [
    "calibrate.py",
    "circuitExec.py",
    "errorCorrection.py",
    "externalScan.py",
    "plot.py",
    "sequence.py",
    "workflow.py",
    "parameters.json",
]

copytree(
    os.path.join(origin, "entropynodes"),
    os.path.join(destination, "entropynodes"),
    dirs_exist_ok=True,
)

for f in file_list:
    copy(os.path.join(origin, f), destination)


print("\t Execute workflow with\tpython -m entropylab.flame.execute ")
print("\t (to interrupt and terminate execution press Ctrl+C)")
print("\n")



import os
import shutil
from datetime import datetime
from pathlib import Path

import pytest

from entropylab.logger import logger
from entropylab.pipeline.results_backend.sqlalchemy.db_initializer import (
    _ENTROPY_DIRNAME,
    _DB_FILENAME,
)

"""conftest.py is a standard pytest configuration file (see here:
https://docs.pytest.org/en/6.2.x/fixture.html).

This conftest.py file contains pytest test fixtures that generate project-related test
data and, on test tear-down, remove the data from disk.
"""


# Test fixtures


@pytest.fixture()
def project_dir_path(request) -> str:
    """Generate a unique path to a (non-existent) project directory
    The directory will be removed recursively on tear down"""
    dir_path = _build_project_dir_path_for_test(request)
    yield dir_path
    _delete_if_exists(dir_path)


@pytest.fixture()
def db_file_path(request) -> str:
    """Generate a unique path to a (non-existent) sqlite db file
    The file will be removed on tear down"""
    dir_path = _build_project_dir_path_for_test(request)
    file_path = dir_path + ".db"
    yield file_path
    _delete_if_exists(file_path)
    _delete_if_exists(dir_path)


@pytest.fixture()
def tinydb_file_path(request) -> str:
    """Generate a unique path to a (non-existent) tinydb json file
    The file will be removed on tear down"""
    dir_path = _build_project_dir_path_for_test(request)
    file_path = os.path.join(dir_path + "/" + "tiny_db.json")
    yield file_path
    _delete_if_exists(dir_path)


@pytest.fixture()
def initialized_project_dir_path(request, project_dir_path) -> str:
    """Create an initialized project and return the path to its directory
    The directory will be removed recursively on tear down"""
    entropy_dir = os.path.join(project_dir_path, _ENTROPY_DIRNAME)
    os.makedirs(entropy_dir)
    # if used in a parametrized test, consider the parameter as a db_template
    # and use a copy of it as the db:
    if hasattr(request, "param") and request.param is not None:
        # param input from test annotation is the db_template file name:
        db_template = os.path.join("./db_templates", str(request.param))
        db_file = os.path.join(entropy_dir, _DB_FILENAME)
        _copy_template(db_template, db_file, request)
    return project_dir_path


# Internal implementations


def _build_project_dir_path_for_test(request):
    project_dir_path = (
        f"tests_cache/{request.node.name}_{datetime.now():%Y-%m-%d-%H-%M-%S}"
    )
    if os.path.isdir(project_dir_path):
        logger.info(f"Test cleanup. Deleting directory '{project_dir_path}'")
        shutil.rmtree(project_dir_path)
    Path(project_dir_path).mkdir(parents=True, exist_ok=True)
    return project_dir_path


def _delete_if_exists(path: str):
    if os.path.isdir(path):
        logger.debug(f"Test cleanup. Deleting directory '{path}'")
        shutil.rmtree(path)
    elif os.path.isfile(path):
        logger.debug(f"Test cleanup. Deleting file '{path}'")
        os.remove(path)


def _copy_template(src, dst, request):
    """Copy the source template file (path relative to test file) to the destination"""
    if src is not None and src != "":
        abs_src = os.path.join(request.fspath.dirname, src)
        logger.debug(f"Copied db template from '{abs_src}' to '{dst}'")
        shutil.copyfile(abs_src, dst)



from .inputs import Inputs
from .outputs import Outputs
import os
import signal
import json
import __main__
import subprocess
import argparse
import redis
import time
import zmq
import msgpack
import platform
from . import nodeio_context

__all__ = [
    "Inputs",
    "Outputs",
    "register",
    "terminate_workflow",
]


def context(name="", description="", icon=""):
    """Description of the current node in the wider context.

    :param name: Name that will be used for providing instances of this node
      in the workflow. It should start with upper case and not contain
      spaces or special characters. E.g. valid name is ExampleName
    :param description: Description of functionality of the node. It will be
      used as documentation of the node.
    :param icon: Optional, icon to use. Icons have to be available in
      entropy_web_ui/nodeicons. Use e.g. 'bootstrap/alarm.svg'
    """

    global status
    global node_name
    global node_description
    global node_dependancies
    global entropy_identity

    parser = argparse.ArgumentParser(prog="Entropy node")
    parser.add_argument("--entropy-identity", type=str, default=None)
    parser.add_argument("--entropy-playbook", type=str, default=None)
    parser.add_argument("args", nargs=argparse.REMAINDER)
    args, _unknown = parser.parse_known_args()

    if args.entropy_identity is not None:
        print("=" * 40)
        print("|- Entropy lab context")
        print(f"|   Node identity:   #{args.entropy_identity}")
        nodeio_context.entropy_identity = args.entropy_identity
    if args.entropy_playbook is not None:
        print(f"|   Playbook:        {args.entropy_playbook}")
        playbook_location = args.entropy_playbook.split(",")
        nodeio_context.playbook = redis.Redis(
            host=playbook_location[0],
            port=playbook_location[1],
            db=playbook_location[2],
        )
        input_values = json.loads(
            nodeio_context.playbook.get(f"#{args.entropy_identity}").decode()
        )
        print("=" * 40)

        external_inputs = Inputs()
        for key, value in input_values.items():
            # Note: type of the variable is not important now, it will be overwritten
            # by the node. We will set initiall all inputs to be states.
            external_inputs.state(key)
            # resolve relative refences for communication if needed
            if type(value) == str and len(value) > 0 and value[0] == "#":
                input_values[key] = "#" + nodeio_context.playbook.get(value).decode()
        external_inputs.set(**input_values)
        Inputs._set_as_external_input(external_inputs)

        nodeio_context.data_server = (
            nodeio_context.playbook.get("dataserver").decode().split(",")
        )
        nodeio_context.runtime_db_connect()
        nodeio_context.node_icon = icon

        zmq_context = nodeio_context.zmq_context()

        executor_input_address = nodeio_context.playbook.get("executor_input").decode()
        socket = zmq_context.socket(zmq.PUB)
        socket.setsockopt(zmq.LINGER, 0)
        socket.setsockopt(zmq.IMMEDIATE, 1)
        socket.connect(executor_input_address)
        nodeio_context.executor_input = socket

        executor_output_address = nodeio_context.playbook.get(
            "executor_output"
        ).decode()
        socket = zmq_context.socket(zmq.SUB)
        socket.setsockopt(zmq.LINGER, 0)
        socket.setsockopt(zmq.IMMEDIATE, 1)
        socket.connect(executor_output_address)
        socket.subscribe("")
        nodeio_context.executor_output = socket

    nodeio_context.node_name = name
    nodeio_context.node_description = description
    nodeio_context.node_icon = icon
    status = StateMachine()


class StateMachine:
    active = True

    def __init__(self):
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def exit_gracefully(self, *args):
        self.active = False
        exit()

    # add status to be executable property; and allow blocking on
    # status check in case we are resolving calibration of the system


def _is_IPython():
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def register():
    """Saves this into node library, allowing it to be used in workflows."""

    if nodeio_context.playbook is not None:
        # wait to establish connections
        upstream_connections = {}
        upstream_data = {}
        outstanding_connections = 0

        for key, value in Inputs._connections().items():
            outstanding_connections += 1
            upstream_connections[key] = value
            upstream_data[key] = None

        outstanding_connections += 1
        upstream_connections["executor_output#"] = nodeio_context.executor_output
        upstream_data["executor_output#"] = None

        while (
            upstream_data["executor_output#"] is None
            or upstream_data["executor_output#"] == ""
        ):
            # send connection ping on all outputs
            for _, out in Outputs._connections().items():
                out.send(msgpack.packb(""))
            # wait
            time.sleep(0.1)
            # non blocking connection ping receive on all inputs
            for key, conn in upstream_connections.items():
                try:
                    a = msgpack.unpackb(conn.recv(flags=zmq.NOBLOCK))
                    if upstream_data[key] is None:
                        outstanding_connections -= 1
                    upstream_data[key] = a
                except zmq.error.ZMQError as e:
                    if e.errno == zmq.EAGAIN:
                        pass
                    else:
                        raise Exception(repr(e))

            # if all incoming connections are established ping ready to executor
            if outstanding_connections == 0:
                registration = {
                    "eui": nodeio_context.entropy_identity,
                    "status": "connected",
                }
                nodeio_context.executor_input.send(msgpack.packb(registration))

            # if executor have not send stop repeat from ping on outputs

        # if executor send stop, stop and ping back you stopped
        registration = {
            "eui": nodeio_context.entropy_identity,
            "status": "waiting_flush",
        }
        nodeio_context.executor_input.send(msgpack.packb(registration))

        # wait for executor to send flush buffer command
        executor = msgpack.unpackb(nodeio_context.executor_output.recv())
        if executor["cmd"] != "flush":
            raise ValueError(
                "Terminating... received cmd from executor ", executor["cmd"]
            )
        # flush and ping back to executor you are ready
        for _key, conn in upstream_connections.items():
            buffer_full = True
            while buffer_full:
                try:
                    a = conn.recv(flags=zmq.NOBLOCK)
                except zmq.error.ZMQError as e:
                    if e.errno == zmq.EAGAIN:
                        buffer_full = False
                    else:
                        raise Exception(repr(e))

        # wait for command to start execution now
        registration = {"eui": nodeio_context.entropy_identity, "status": "ready"}
        nodeio_context.executor_input.send(msgpack.packb(registration))

        # block untill all nodes are up
        executor = msgpack.unpackb(nodeio_context.executor_output.recv())
        if executor["cmd"] != "start":
            raise ValueError(
                "Terminating... received cmd from executor ", executor["cmd"]
            )
        # TO-DO - check that saved node schema is the same as the current node
        # schema (i.e. node code has not been changed since the last time we
        # indenpendantly run this node)

        return  # We are running as a part of the workflow. Skip node creation.

    if _is_IPython():
        return  # this node is still being prepared for runtime

    bin_path = str(os.path.basename(__main__.__file__))

    schema = {
        "name": nodeio_context.node_name,
        "description": nodeio_context.node_description,
        "command": "python3",
        "bin": bin_path,
        "icon": nodeio_context.node_icon,
        "inputs": Inputs._schema(),
        "outputs": Outputs._schema(),
    }

    destination = os.path.join(os.path.join(os.getcwd(), "entropynodes"), "schema")
    if not os.path.exists(destination):
        os.makedirs(destination)
    with open(
        os.path.join(destination, f"{nodeio_context.node_name}.json"),
        encoding="utf-8",
        mode="w",
    ) as f:
        f.write(json.dumps(schema))

    env = os.environ
    env["PYTHONPATH"] = os.getcwd()

    if platform.system() == "Windows":
        python_cmd = "python"
    else:
        python_cmd = "python3"

    subprocess.run(
        f"{python_cmd} -m entropylab.flame.generate_node {nodeio_context.node_name}",
        env=env,
        shell=True,
        universal_newlines=True,
        start_new_session=True,
    )


def terminate_workflow():
    if nodeio_context.playbook is not None:
        # request executor to terminate
        msg = {
            "eui": nodeio_context.entropy_identity,
            "status": "requests workflow termination",
        }
        nodeio_context.executor_input.send(msgpack.packb(msg))
        nodeio_context.runtime_db_close()
        nodeio_context.zmq_context().term()
        status.active = False
        exit()
    else:
        # terminate just this node
        status.active = False
        exit()


def terminate_node():
    if _is_IPython():

        class StopExecution(Exception):
            def _render_traceback_(self):
                pass

        raise StopExecution
    else:
        exit()



import atexit
import os
import json
import networkx as nx
from .enum_types import ioFormat

__all__ = ["Workflow"]


class BagOfNodes(object):
    def __init__(self):
        pass

    def __repr__(self) -> str:
        """node list"""
        node_list = ""
        for a in dir(self):
            if not a.startswith("__"):
                node_list += f"\n{a}"
        return node_list


class NodeGroup(object):
    def __init__(self, name):
        self.name = name
        self.nodes = []

    def add(self, node):
        for x in self.nodes:
            if x._name == node._name:
                raise ValueError(
                    f"Node '{node._name}' is already added to group '{self.name}'"
                )
        self.nodes.append(node)


class Workflow(object):

    __workflows = []
    __current_workflow_context = None

    @classmethod
    def _register_node(cls, node_instance, prepend_name=""):
        if cls.__current_workflow_context is None:
            raise ValueError(
                "Before creating instances of nodes you have"
                " to define Workflow instance"
            )
        node_list = cls.__workflows[cls.__current_workflow_context]._nodes
        node_name = prepend_name + node_instance._name
        if node_name in node_list:
            raise ValueError(
                ("Node with Entropy Unique Identifier '%s'" % node_name)
                + " already exists!\nPlease make sure node names are unique.\n"
                + "Aborting workflow creation"
            )
        node_list[node_name] = node_instance
        setattr(
            cls.__workflows[cls.__current_workflow_context].n, node_name, node_instance
        )

    @classmethod
    def _add_to_workflow_list(cls, workflow_instance):
        cls.__workflows.append(workflow_instance)
        atexit.register(workflow_instance._generate_parameter_file)
        return len(cls.__workflows) - 1

    @classmethod
    def _main_workflow(cls):
        if cls.__current_workflow_context is None:
            return None
        return cls.__workflows[cls.__current_workflow_context]

    @classmethod
    def _main_workflow_index(cls):
        return cls.__current_workflow_context

    @classmethod
    def _register_main_workflow(cls, workflow_instance):
        cls.__current_workflow_context = workflow_instance._index

    @classmethod
    def _prevent_parameter_file_overwrite(cls):
        cls.__current_workflow_context = None

    def __init__(self, name, description=""):
        """New workflow

        :param name: name of workflow
        :param description: longer description of workflow
        """
        self._nodes = {}
        self.name = name
        self.description = description
        #: """ Access nodes by names as workflow.n.<node_name> """
        self.n = BagOfNodes()
        self._index = self._add_to_workflow_list(self)
        self._views = {}
        self._node_info = {}  #: type of nodes used
        self.register()

    def register(self):
        """
        When workflow is registred, if workflow file is executed, the
        workflow will generate parameter list. This workflow file can also
        then be used for executor
        """
        self._register_main_workflow(self)

    def _generate_parameter_file(self):
        if self._index != self._main_workflow_index():
            return
        full_path_out = os.path.join(os.getcwd(), "parameters.json")
        existing_parameters = None
        if os.path.exists(full_path_out):
            # load existing file
            try:
                with open(full_path_out, "r") as file:
                    existing_parameters = json.load(file)
            except Exception:
                pass
        with open(full_path_out, encoding="utf-8", mode="w") as f:
            f.write(self._list_missing_parameters())

        if existing_parameters is not None:
            # use previously existing parameters.json to resolve values
            try:
                reused_values = 0
                with open(full_path_out, "r") as file:
                    parameters = json.load(file)
                for node, values in existing_parameters.items():
                    if node in parameters:
                        for key, value in values.items():
                            if key in parameters[node]:
                                if value is not None:
                                    parameters[node][key] = value
                                    reused_values += 1
                if reused_values > 0:
                    with open(full_path_out, encoding="utf-8", mode="w") as f:
                        f.write(json.dumps(parameters, indent=2))
                    manyvalues = "s" if reused_values > 1 else ""
                    print(
                        f"Reused {reused_values} parameter value{manyvalues} "
                        "from existing parameters.json file."
                    )
            except Exception:
                pass

    def _resolve_parameters(self, json_filename):
        with open(json_filename, "r") as file:
            parameters = json.load(file)

        for node, node_param in parameters.items():
            self._nodes[node]._inputs._inputs.set(**node_param)
        for node_name, node in self._nodes.items():
            if not node._inputs._inputs.completely_defined():
                print(f"\tNode '{node_name}' parameters are not completely resolved!")
                print("\tMissing parameters:")
                print(self._list_missing_parameters())
                exit()
                # raise ValueError("Execution cannot run before all parameters are resolved.")

    def _list_missing_parameters(self, existing_parameters=None):
        parameter_listing = "{\n"
        unresolved_nodes = 0
        for name, node in self._nodes.items():
            if not node._inputs._inputs.completely_defined():
                if unresolved_nodes != 0:
                    parameter_listing += ",\n"
                parameter_listing += (
                    f'"{name}":' + node._inputs._inputs.print_undefined()
                )
                unresolved_nodes += 1
        parameter_listing += "\n}"
        return parameter_listing

    def _resolved_parameters(self):
        parameters = {}
        for name, node in self._nodes.items():
            parameters[name] = node._inputs._inputs.get_as_JSON(
                what="defined", format=ioFormat.VALUES
            )
        return parameters

    def _to_json(self):
        """Export as JSON with Cytoscape graph representation"""
        graph = nx.DiGraph()
        icons = {}
        # add nodes
        for node_name, node in self._nodes.items():
            graph.add_node(node_name, ident=node_name, name=type(node).__name__)
            icons[type(node).__name__] = node._icon
        # add links
        for node_name, node in self._nodes.items():
            for _input_name, input_value in node._inputs._inputs.values.items():
                if (
                    input_value is not None
                    and type(input_value[0]) == str
                    and len(input_value[0]) > 2
                    and input_value[0][0] == "#"
                    and input_value[0][1] != "/"
                ):
                    # relative EUI reference to runtime variable
                    source_node = input_value[0][1:].split("/")[0]
                    graph.add_edge(source_node, node_name)

        result = {
            "name": self.name,
            "description": self.description,
            "graph": nx.cytoscape_data(graph),
            "resolved_inputs": self._resolved_parameters(),
            "icons": icons,
        }

        return json.dumps(result)

    def _node_details(self, node):
        node_class = type(node).__name__
        if node_class not in self._node_info:
            json_filename = os.path.join(
                os.getcwd(), "entropynodes", "schema", f"{node_class}.json"
            )
            with open(json_filename, "r") as file:
                self._node_info[node_class] = json.load(file)
        return self._node_info[node_class]

    def _summary_json(self):
        """Export as JSON summary of indexable information about workflow"""

        used_node_types = {}
        for _, node in self._nodes.items():
            node_info = self._node_details(node)
            node_class = type(node).__name__

            inputs = {}
            for input_name, input in node_info["inputs"][0]["description"].items():
                units = node_info["inputs"][0]["units"][input_name]
                inputs[input_name + f" ({units})"] = input
            outputs = {}
            for output_name, output in node_info["outputs"][0]["description"].items():
                units = node_info["outputs"][0]["units"][output_name]
                outputs[output_name + f" ({units})"] = output
            outputs_retention = {}
            for output_name, retention in node_info["outputs"][0]["retention"].items():
                outputs_retention[output_name] = retention
            used_node_types[node_class] = {
                "description": node_info["description"],
                "inputs": inputs,
                "outputs": outputs,
                "outputs_retention": outputs_retention,
            }

        result = {
            "name": self.name,
            "description": self.description,
            "nodes": used_node_types,
        }

        return json.dumps(result)

    def add(self, node_or_workflow, prepend_name=""):
        if isinstance(node_or_workflow, Workflow):
            for _name, node in node_or_workflow._nodes.items():
                self._register_node(node, prepend_name=prepend_name)
        else:
            self._register_node(node_or_workflow)

    def parameters(self):
        pass

    def group(self, node_list, group_name, view="default"):
        pass



import entropylab.flame.nodeio as nodeio
import time
import random

# ==================== DEFINE NODE ====================

nodeio.context(
    name="QPUcircuitRunner",
    description="Executes given circuit sequence",
    icon="bootstrap/cpu.svg",
)

input = nodeio.Inputs()
input.stream("circuit_param", units="JSON", description="cirucuit description")
input.stream("error_correction", units="JSON", description="corrected data")

output = nodeio.Outputs()
output.define(
    "precorrected_data",
    units="JSON",
    description="data send for error correction",
    retention=2,
)
output.define(
    "circuit_finished", units="bool", description="Event trigger", retention=0
)
output.define(
    "final_data", units="bool", description="circuit averaged output", retention=2
)

nodeio.register()  # enables this to be used as part of workflow

# ==================== DRY-RUN DATA ====================

input.set(circuit_param=7.3)
input.set(
    error_correction=[[2, 1], [1, 2]],
)

# =============== RUN NODE STATE MACHINE ===============

while nodeio.status.active:
    circuit = input.get("circuit_param")
    print(f"running circuit param : {circuit}")
    time.sleep(0.1)

    output.set(
        precorrected_data=[
            [circuit + random.random() * 10, circuit / 2 + random.random() * 10],
            [circuit / 2 + random.random() * 10, circuit + random.random() * 10],
        ]
    )
    errorCorrected = input.get("error_correction")
    output.set(final_data=errorCorrected[0][0])
    output.set(circuit_finished=True)

# ================= =================== =================



from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Optional, Type
from warnings import warn

from bokeh.models import Renderer
from bokeh.plotting import Figure
from matplotlib.figure import Figure as matplotlibFigure
from plotly import graph_objects as go


@dataclass
class ExperimentInitialData:
    """
    Information about the experiment that is saved on start
    """

    label: str
    user: str
    lab_topology: str
    script: str
    start_time: datetime
    story: str = None


@dataclass
class ExperimentEndData:
    """
    Information about the experiment that is saved on start
    """

    end_time: datetime
    success: bool


@dataclass
class RawResultData:
    """
    A single result that will be saved
    """

    label: str
    data: Any
    stage: int = -1
    story: str = None

    def __repr__(self):
        return f"<RawResultData(stage='{self.stage}', label='{self.label}')>"


@dataclass
class Metadata:
    """
    A single metadata that will be saved
    """

    label: str
    stage: int
    data: Any


@dataclass
class Debug:
    """
    extra information about the experiment execution for better debugging
    """

    python_env: str
    python_history: str
    station_specs: str
    extra: str


class PlotGenerator(ABC):
    """
    An abstract class for plots.
    Implementations of this class will let Entropy to save and view plots.
    Every implementation can either implement all plotting functions
    (within the different environments), or just part of it.
    """

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def plot_bokeh(self, figure: Figure, data, **kwargs) -> Renderer:
        """
            plot the given data within the Bokeh Figure
        :param figure: Bokeh figure to plot in
        :param data: plot data
        :param kwargs: extra parameters for plotting
        """
        pass

    @abstractmethod
    def plot_matplotlib(self, figure: matplotlibFigure, data, **kwargs):
        """
            plot the given data within the matplotlib Figure
        :param figure: matplotlib figure
        :param data: plot data
        :param kwargs: extra parameters for plotting
        """
        pass

    @abstractmethod
    def plot_plotly(self, figure: go.Figure, data, **kwargs) -> None:
        """
            plot the given data within the plot.ly Figure
        :param figure: plot.ly figure
        :param data: plot data
        :param kwargs: extra parameters for plotting
        """
        pass


@dataclass(frozen=True, eq=True)
class PlotSpec:
    """
    Description and plotting instructions for a plot that will be saved
    """

    generator: Optional[Type[PlotGenerator]] = None
    label: Optional[str] = None
    story: Optional[str] = ""


@dataclass
class NodeData:
    """
    information about a specific node
    """

    stage_id: int
    start_time: datetime
    label: str
    is_key_node: bool


class DataWriter(ABC):
    """
    An abstract class for Entropy database, defines the way entropy saves data
    of experiments.
    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def save_experiment_initial_data(self, initial_data: ExperimentInitialData) -> int:
        """
        save experiment information to db according to the ExperimentInitialData class
        """
        pass

    @abstractmethod
    def save_experiment_end_data(self, experiment_id: int, end_data: ExperimentEndData):
        """
        save experiment information to db according to the ExperimentEndData class
        """
        pass

    @abstractmethod
    def save_result(self, experiment_id: int, result: RawResultData):
        """
        save a new result to the db according to the RawResultData class
        """
        pass

    @abstractmethod
    def save_metadata(self, experiment_id: int, metadata: Metadata):
        """
        save a new metadata to the db according to the Metadata class
        """
        pass

    @abstractmethod
    def save_debug(self, experiment_id: int, debug: Debug):
        """
        save experiment debug information to the db according to the Debug class
        """
        pass

    @abstractmethod
    def save_plot(self, experiment_id: int, plot: PlotSpec, data: Any):
        """
            save a new plot to the db according to the PlotSpec class
        :param experiment_id: the experiment id
        :param plot: plotting instructions
        :param data: the data of the plot
        """
        warn(
            "This method will soon be deprecated. Please use save_figure() instead",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        pass

    def save_figure(self, experiment_id: int, figure: go.Figure) -> None:
        """
            save a new plotly figure to the db and associates it with an experiment

        :param experiment_id: the id of the experiment to associate the figure to
        :param figure: the figure to save to the database
        """
        pass

    @abstractmethod
    def save_node(self, experiment_id: int, node_data: NodeData):
        """
        saves graph's node data to the db, according to NodeData class
        """
        pass



# Flame executor

import os
import sys
import time
import json
import psutil
import importlib
import platform

import zmq
import h5py
import msgpack

from sqlalchemy import text as sql_text

from entropylab.flame.workflow import Workflow
from entropylab.flame.execute import _utils as execute_utils
from entropylab.flame.execute._config import _Config, logger
from entropylab.flame.execute._message_queue_info import MessageQueueInfo
from entropylab.flame.execute._runtime_state_info import RuntimeStateInfo


class Execute:
    def __init__(self, args_):
        self.args = args_
        self.executor_input = None
        self.executor_output = None
        self.port_input = None
        self.port_output = None
        self.pid2node = {}
        self.node2process = {}
        self.processes_list = []
        self.cwd = os.getcwd()
        self.metadata = json.loads(self.args.metadata)
        self.__init_workflow()

        _Config.job_id = self.metadata.get("job_eui", "#/").replace("#/", "")
        if _Config.job_id == "":
            # for the Flame executed from command line without job EUI
            _Config.job_id = "output_data"
        runtime_id = self.metadata.get("runtime_id", -1)
        self.routing_key = f"status_updates.{runtime_id}.{_Config.job_id}"
        self.runtime_state_info = RuntimeStateInfo()
        self.message_queue_info = MessageQueueInfo()

    def __init_workflow(self):
        logger.info("Execute. Init workflow")
        workflow_path = self.args.workflow.replace(".py", "")
        importlib.import_module(workflow_path)
        self.workflow = Workflow._main_workflow()
        # now prevent workflow to overwrite parameters on exit
        Workflow._prevent_parameter_file_overwrite()

        self.workflow._resolve_parameters(self.args.parameters)

    def __wait_for_processes(
        self,
        additional_process_list,
        message: str,
        style: str,
        wait_process_timeout: float = 3,
    ):
        logger.debug(
            f"Execute. Wait for processes. "
            f"Processes list: {self.processes_list}; "
            f"additional list: {additional_process_list}; "
            f"message: {message}; style: {style}; "
            f"wait timeout: {wait_process_timeout}."
        )
        gone, alive = psutil.wait_procs(
            self.processes_list + additional_process_list, timeout=wait_process_timeout
        )
        for g in gone:
            if g.pid in self.pid2node:
                node_name = self.pid2node[g.pid]
                execute_utils.status_update(
                    node_name,
                    self.routing_key,
                    message,
                    style,
                    self.message_queue_info.updates_channel,
                )
                code = g.returncode
                if code is not None and code != 0 and code != 15:
                    # code == 15 is for forceful termination since on Windows
                    # there is no graceful termination signal
                    # see https://bugs.python.org/issue26350
                    execute_utils.status_update(
                        node_name,
                        self.routing_key,
                        f"error, exit code {code}",
                        "error",
                        self.message_queue_info.updates_channel,
                    )
                self.processes_list.remove(g)
        return gone, alive

    def __setup_executor_input_output(self, port_number=9000):
        logger.debug(
            f"Execute. Setup executor input and output. Port number: {port_number}"
        )
        # open port for executor in and out channel
        zmq_context = zmq.Context()
        runtime_state = self.runtime_state_info.runtime_state
        port_number = execute_utils.get_free_port(
            port_number, runtime_state, "executor_input"
        )
        logger.debug(
            f"Execute. Setup executor input and output. "
            f"Input port number: {port_number}."
        )
        self.port_input = port_number
        port_address = f"tcp://127.0.0.1:{port_number}"
        runtime_state.set("executor_input", port_address)

        self.executor_input = zmq_context.socket(zmq.SUB)
        self.executor_input.setsockopt(zmq.LINGER, 0)
        self.executor_input.setsockopt(zmq.RCVTIMEO, 200)  # 0.2 s timeout
        self.executor_input.bind(port_address)  # receives from many publishers
        self.executor_input.subscribe("")

        port_number = execute_utils.get_free_port(
            port_number, runtime_state, "executor_output"
        )
        logger.debug(
            f"Execute. Setup executor input and output. "
            f"Output port number: {port_number}"
        )
        port_address = f"tcp://127.0.0.1:{port_number}"
        runtime_state.set("executor_output", port_address)
        self.port_output = port_number

        self.executor_output = zmq_context.socket(zmq.PUB)
        self.executor_output.setsockopt(zmq.LINGER, 0)
        self.executor_output.bind(port_address)  # sends to many publishers

    def __zmq_messages_status_check(
        self,
        total_node_count,
        connection_wait,
        status_check,
        need_send_output_msg=False,
    ):
        logger.debug(
            f"Execute. ZMQ messages status check. "
            f"Node count: {total_node_count}; "
            f"connection wait: {connection_wait}; "
            f"Status check: {status_check};"
            f"Need send output: {need_send_output_msg}."
        )
        ready_nodes_count = 0
        start_time_ = time.time()
        while (
            ready_nodes_count < total_node_count
            and _Config.execution_active
            and time.time() - start_time_ < connection_wait
        ):
            if need_send_output_msg:
                self.executor_output.send(msgpack.packb(""))

            try:
                update_ = msgpack.unpackb(self.executor_input.recv())
                if _Config.node_status_dict[update_["eui"]] != update_["status"]:
                    _Config.node_status_dict[update_["eui"]] = update_["status"]
                    if update_["status"] == status_check:
                        ready_nodes_count += 1
            except zmq.error.ZMQError as e:
                if e.errno == zmq.EAGAIN:
                    pass
        logger.debug(
            f"Execute. ZMQ messages status check. "
            f"Ready nodes count: {ready_nodes_count}"
        )
        return ready_nodes_count

    def __init_nodes(self, node_schemas_):
        # initialize all the nodes, saving process id
        logger.debug(f"Execute. Init nodes. Node schemas: {node_schemas_}.")
        log_directory = os.path.join(self.cwd, "entropylogs")
        playbook_server_arg = (
            f"{self.runtime_state_info.playbook_server[0]},"
            f"{self.runtime_state_info.playbook_server[1]},"
            f"{self.runtime_state_info.playbook_server[2]}"
        )
        if not os.path.exists(log_directory):
            os.makedirs(log_directory)
        try:
            for node_name, node in self.workflow._nodes.items():
                node_class = type(node).__name__
                schema = node_schemas_[node_class]
                log_file = os.path.join(log_directory, f"{node_name}.txt")
                if platform.system() == "Windows" and (
                    schema["command"] == "python" or schema["command"] == "python3"
                ):
                    # -u is to unbuffer stoud and sterr and to write directly to log file
                    exec_command = "python -u"
                else:
                    exec_command = schema["command"]
                cmd = (
                    f'{exec_command} {os.path.join(self.cwd, schema["bin"])}'
                    + f' --entropy-identity "{node_name}"'
                    + f' --entropy-playbook "{playbook_server_arg}"'
                    + f' > "{log_file}" 2>&1'
                )
                p = psutil.Popen(
                    cmd, shell=True, universal_newlines=True, start_new_session=True
                )

                self.node2process[node_name] = p
                self.pid2node[p.pid] = node_name
                self.processes_list.append(p)
                self.runtime_state_info.runtime_state.set(f"pid#{node_name}", p.pid)

                execute_utils.status_update(
                    node_name,
                    self.routing_key,
                    "initialised",
                    "initialised",
                    self.message_queue_info.updates_channel,
                )

        except OSError as e:
            # the system failed to execute the shell
            # (out-of-memory, out-of-file-descriptors, and other extreme cases)
            logger.error(
                f"Execute. Init nodes. Fail to run shell.\n"
                f"etails: {str(e)}.\n"
                f"Traceback: {e.__traceback__}"
            )
            sys.exit("failed to run shell: '%s'" % (str(e)))

    def __assign_relative_outputs_to_free_ports(self):
        logger.debug("Execute. Assign relative outputs to free ports.")
        total_node_count = 0
        resolution = {}
        node_schemas = {}
        with execute_utils.get_runtimedata(_Config.DATABASE_NAME) as db:
            # check all relative outputs and assign them free ports
            # write resolution in playbook
            # TODO: if using UNIX compatible system,
            # and Windows 10 (https://github.com/zeromq/libzmq/issues/3691)
            # use ipc instead of tcp to reduce latency

            for node_name, node in self.workflow._nodes.items():
                total_node_count += 1
                node_info = self.workflow._node_details(node)
                for o in node._outputs._outputs:
                    eui = f"#{node_name}/{o}"

                    if node_info["outputs"][0]["retention"][o] > 0:
                        # store permanently this value
                        # create space in runtime database for storing this output
                        with db.begin():
                            db.execute(
                                sql_text(
                                    f"""
                                    CREATE TABLE "{eui}" (
                                    time TIMESTAMPTZ NOT NULL,
                                    value JSONB
                                    );
                                    """
                                )
                            )
                            db.execute(
                                sql_text(
                                    f"""SELECT create_hypertable('"{eui}"', 'time');"""
                                )
                            )

                    port_number = execute_utils.get_free_port(
                        _Config.port_number, self.runtime_state_info.runtime_state, eui
                    )

                    port_address = f"tcp://127.0.0.1:{port_number}"
                    resolution[eui] = port_address
                    self.runtime_state_info.runtime_state.set(eui, port_address)

                    node_class = type(node).__name__
                    if node_class not in node_schemas:
                        full_path_in = os.path.join(
                            os.path.join(self.cwd, "entropynodes"), "schema"
                        )
                        full_path_in = os.path.join(full_path_in, f"{node_class}.json")
                        with open(full_path_in, encoding="utf-8", mode="r") as f:
                            schema = json.loads(f.read())
                        node_schemas[node_class] = schema
                _Config.node_status_dict[node_name] = "resolved"
        logger.debug(
            f"Execute. Assign relative outputs to free ports."
            f"Node count: {total_node_count}; "
            f"Node schemas: {node_schemas}."
        )
        return total_node_count, node_schemas

    def __write_parameter_resolution_in_the_playbook(self):
        logger.debug("Execute. Write parameter resolution in the playbook.")
        for node_name, node in self.workflow._nodes.items():
            self.runtime_state_info.runtime_state.set(
                f"#{node_name}", node._inputs._inputs.print_defined()
            )

    def __start_execution_if_ready(self, ready_nodes_count, total_node_count):
        logger.debug(
            f"Execute. Start execution if ready. "
            f"Ready nodes count: {ready_nodes_count}. "
            f"Nodes count: {total_node_count}."
        )
        if ready_nodes_count == total_node_count:
            # tell nodes to stop sending on outputs and wait buffer flush
            self.executor_output.send(msgpack.packb({"cmd": "wait_flush"}))

            # wait for all nodes to stop
            self.__zmq_messages_status_check(
                total_node_count,
                self.args.connection_wait,
                "waiting_flush",
            )

            # flush buffer outputs
            self.executor_output.send(msgpack.packb({"cmd": "flush"}))

            # wait for all nodes to flush outputs
            self.__zmq_messages_status_check(
                total_node_count,
                self.args.connection_wait,
                "ready",
            )

            # start execution
            execution_connection_problem = False
            self.executor_output.send(msgpack.packb({"cmd": "start"}))
        else:
            execution_connection_problem = True
            _Config.execution_active = False
        logger.debug(
            f"Execute. Start execution if ready. "
            f"Execution connection problem: {execution_connection_problem}."
        )
        return execution_connection_problem

    def __wait_for_finish_or_terminate(self):
        # wait for finish, if debug on periodically snapshot process performance
        # also listen to external signal (in case the execution termination request)
        # is received
        logger.debug("Execute. Wait for finish or terminate.")
        max_time = self.args.max_execution_time
        active_nodes = len(self.processes_list)
        time_step = self.args.status_check_interval

        while (
            _Config.execution_active
            and (max_time > 0 or (self.args.max_execution_time == 0))
            and active_nodes > 0
        ):
            gone, alive = self.__wait_for_processes(
                [],
                "success",
                "finished",
                wait_process_timeout=time_step,
            )
            active_nodes = len(alive)
            execute_utils.check_node_messages(self.executor_input)
            max_time -= time_step
        execution_timeout_event = not (
            max_time > 0 or (self.args.max_execution_time == 0)
        )
        logger.debug(
            f"Execute. Wait for finish or terminate."
            f"Execution timeout event: {execution_timeout_event}"
        )
        return execution_timeout_event

    def __terminate(self):
        logger.debug("Execute. Terminate.")
        # send terminate signal to all the nodes
        processListAll = []
        for p in self.processes_list:
            for child in p.children(recursive=True):
                processListAll.append(child)
        for p in processListAll:
            p.terminate()

        # wait for nodes to gracefully exit
        logger.debug("Execute. Terminate. Wait for nodes to gracefully exit. v0")
        gone, alive = self.__wait_for_processes(processListAll, "success", "finished")

        # kill all the remaining nodes
        logger.debug("Execute. Terminate. Kill all the remaining nodes. v0")
        for p in alive:
            if p not in self.processes_list:
                # kill only executed process, not the shell script itself
                # in order to keep log file
                p.kill()

        logger.debug("Execute. Terminate. Wait for finish. v1")
        gone, alive2 = self.__wait_for_processes(alive, "", "finished")

        logger.debug("Execute. Terminate. Kill all the remaining nodes. v1")
        for p in self.processes_list:
            # kill even shell script itself if it is still active
            p.kill()

        logger.debug("Execute. Terminate. Wait for finish. v2")
        self.__wait_for_processes(alive2, "success", "finished")

    def __flush_runtimedata_into_hdf5(self):
        # add def for decrease tab level
        def _update_node_output(db, node_name, o, node_info, grp):
            eui = f"#{node_name}/{o}"

            if node_info["outputs"][0]["retention"][o] == 2:
                results = (
                    db.execute(sql_text(f'SELECT value FROM "{eui}"')).scalars().all()
                )
                results_time = (
                    db.execute(
                        sql_text(
                            f"""
        SELECT to_char
        (time::timestamp at time zone 'UTC', 'YYYY-MM-DD"T"HH24:MI:SS.MS"Z"')
        FROM "{eui}"
        """
                        )
                    )
                    .scalars()
                    .all()
                )
                dset = grp.create_dataset(f"{o}_time", data=results_time)
                try:
                    dset = grp.create_dataset(o, data=results)
                except TypeError:
                    r = []
                    for elem in results:
                        r.append(json.dumps(elem))
                    dset = grp.create_dataset(o, data=r)
                dset.attrs["description"] = node_info["outputs"][0]["description"][o]
                dset.attrs["units"] = node_info["outputs"][0]["units"][o]

        # add def for decrease tab level
        def _update_node(db, node_name, node, node_info):
            # node_info = self.workflow._node_details(node)
            grp = f.create_group(f"{node_name}")
            grp.attrs["type"] = node_info["name"]
            grp.attrs["description"] = node_info["description"]
            grp.attrs["bin"] = node_info["bin"]

            for o in node._outputs._outputs:
                _update_node_output(db, node_name, o, node_info, grp)

        # Flush the runtimedata into data_store HDF5 file
        logger.debug("Execute. Flush runtime data into hdf5.")
        with execute_utils.get_runtimedata(_Config.DATABASE_NAME) as db:
            with h5py.File(f"./{_Config.job_id}.hdf5", "w") as f:
                execute_utils.write_metadata_to_h5(f, self.metadata)
                for node_name, node in self.workflow._nodes.items():
                    node_info = self.workflow._node_details(node)
                    _update_node(db, node_name, node, node_info)

    def __clean_the_playbook(self):
        logger.debug("Execute. Clean the playbook.")
        runtime_state = self.runtime_state_info.runtime_state
        for node_name, node in self.workflow._nodes.items():
            runtime_state.delete(f"pid#{node_name}")
            for o in node._outputs._outputs:
                eui = f"#{node_name}/{o}"
                port_number = runtime_state.get(eui).decode().split(":")[-1]
                runtime_state.delete(f"system/port{port_number}")
                runtime_state.delete(eui)

            runtime_state.delete(f"#{node_name}")

    def __clean(self):
        logger.debug("Execute. Clean.")
        self.runtime_state_info.runtime_state.delete("dataserver")

        if self.message_queue_info.updates_channel is not None:
            self.message_queue_info.status_connection.close()

        self.runtime_state_info.runtime_state.delete("executor_output")
        self.runtime_state_info.runtime_state.delete("executor_input")
        self.executor_input.close()
        self.executor_output.close()
        execute_utils.remove_port_lock(
            self.port_input, self.runtime_state_info.runtime_state
        )
        execute_utils.remove_port_lock(
            self.port_output, self.runtime_state_info.runtime_state
        )

        self.runtime_state_info.runtime_state.delete("executor_pid")

    def __result_message(self, execution_timout_event, execution_connection_problem):
        logger.debug("Execute. Result message.")
        status_message = ""
        if execution_timout_event:
            status_message = "Execution timed out."
        if execution_connection_problem:
            status_message = (
                "Connection between nodes timed out before being established."
            )

        nodes_success = 0
        nodes_error = 0

        for _node, value in _Config.node_status_dict.items():
            if value != "success":
                nodes_error += 1
            else:
                nodes_success += 1
        if nodes_error == len(_Config.node_status_dict):
            status = "Failure"
        elif status_message == "" and nodes_error == 0:
            status = "Success"
        else:  # previous condition: nodes_error > 0.
            status = "Partially successful"

        return json.dumps(
            {
                "status": status,
                "status_message": status_message,
                "nodes_success": nodes_success,
                "nodes_error": nodes_error,
                "nodes_count": len(_Config.node_status_dict),
                "nodes": _Config.node_status_dict,
            }
        )

    def run(self):
        self.__setup_executor_input_output(_Config.port_number)
        total_node_count, node_schemas = self.__assign_relative_outputs_to_free_ports()
        self.__write_parameter_resolution_in_the_playbook()
        self.runtime_state_info.runtime_state.set("executor_pid", os.getpid())

        # initialize all the nodes, saving process id
        self.__init_nodes(node_schemas)

        # make sure all requested nodes are up before giving the green light for
        # them to start communicating
        ready_nodes_count = self.__zmq_messages_status_check(
            total_node_count,
            self.args.connection_wait,
            "connected",
            True,
        )

        execution_connection_problem = self.__start_execution_if_ready(
            ready_nodes_count, total_node_count
        )
        execution_timeout_event = self.__wait_for_finish_or_terminate()
        self.__terminate()
        # clean the playbook
        self.__clean_the_playbook()
        self.__flush_runtimedata_into_hdf5()
        # clean playbook reference to data storage location
        self.__clean()
        result_message = self.__result_message(
            execution_timeout_event, execution_connection_problem
        )
        logger.debug(f"Execute. Run. Result: {result_message}")
        return result_message



# ==================== DEFINE NODE ====================
import entropylab.flame.nodeio as nodeio

nodeio.context(
    name="NameForThisNodeClass",
    description="what it does",
    icon="bootstrap/gear-fill.svg",  # any bootstrap icon, see https://icons.getbootstrap.com/
)

inputs = nodeio.Inputs()
# define inputs
# inputs.stream( "input name", units="units or data type", description="some detailed explanation" )
# inputs.state(  "state input name", units="units or data type", description="some detailed explanation" )

outputs = nodeio.Outputs()
# define outputs
# outputs.define( "output name", units="units or data type", description="some detailed explanation" )

nodeio.register()

# ==================== DRY RUN DATA ====================

# set inputs data for dry-run of the node
inputs.reset_all()  # needed just for repeated for development with repeated runs in IPython
# inputs.set( <input_name> = <input_value>)

# =============== RUN NODE STATE MACHINE ===============

while nodeio.status.active:

    # here write code that solves the problem this node
    # deals with, controlling execution flow using when needed
    # inputs.get( ... )   inputs.updated( ... ) and outputs.set ( ... )
    # you can programmatically terminate whole workflow by calling
    # nodeio.terminate_workflow()
    # or you can exit this individual node by calling usual exit()

    pass  # delete this line when writting your code



import dash_bootstrap_components as dbc
import pkg_resources
from dash import html


def footer():
    version = pkg_resources.get_distribution("entropylab").version

    href_home = "https://github.com/entropy-lab/entropy/"
    href_documentation = "https://entropylab-docs-open.netlify.app/"
    href_feedback = "https://github.com/entropy-lab/entropy/issues/"
    return dbc.Row(
        children=[
            dbc.Col(
                dbc.Row(
                    children=[
                        dbc.Col(f"Entropylab v{version}", width=4),
                        dbc.Col(" · ", width=1),
                        dbc.Col(
                            dbc.Row(
                                children=[
                                    dbc.Col(
                                        html.A(
                                            "Home",
                                            href=href_home,
                                            target="_blank",
                                        ),
                                        width=3,
                                    ),
                                    dbc.Col("|", width=1),
                                    dbc.Col(
                                        html.A(
                                            "Documentation",
                                            href=href_documentation,
                                            target="_blank",
                                        ),
                                        width=4,
                                    ),
                                    dbc.Col("|", width=1),
                                    dbc.Col(
                                        html.A(
                                            "Feedback",
                                            href=href_feedback,
                                            target="_blank",
                                        ),
                                        width=3,
                                    ),
                                ],
                                id="footer-links",
                            ),
                            width=7,
                        ),
                    ]
                ),
                width=dict(size=4, offset=4),
            ),
        ],
        style=dict(marginTop="40px", marginBottom="40px"),
    )



import pika
from pika.exchange_type import ExchangeType

from entropylab.flame.execute._config import logger


def _setup_message_queue(host_url: str = "localhost"):
    logger.debug("MessageQueueInfo. Setup message queue")
    status_connection = pika.BlockingConnection(
        pika.URLParameters(f"amqp://guest:guest@{host_url}:5672")
    )
    updates_channel = status_connection.channel()
    updates_channel.exchange_declare(
        "amq.topic", exchange_type=ExchangeType.topic, durable=True
    )
    return status_connection, updates_channel


class MessageQueueInfo:
    def __init__(self):
        try:
            self.status_connection, self.updates_channel = _setup_message_queue(
                "localhost"
            )
            logger.info("MessageQueueInfo. Host: localhost")
        except Exception:
            try:
                self.status_connection, self.updates_channel = _setup_message_queue(
                    "messagequeue"
                )
                logger.info("MessageQueueInfo. Host: messagequeue")
            except Exception:
                self.status_connection, self.updates_channel = None, None
                logger.info("MessageQueueInfo. No message queue configured.")



from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import List, Any, Optional, Iterable
from warnings import warn

from pandas import DataFrame
from plotly import graph_objects as go

from entropylab.pipeline.api.data_writer import PlotGenerator


class ScriptViewer:
    def __init__(self, stages: List[str]) -> None:
        super().__init__()
        self._stages = stages

    def show_stages(self):
        return "\n".join(self._stages)

    def print_stage(self, i):
        return self._stages[i]

    def print_all(self):
        return "\n".join(self._stages)


@dataclass
class ExperimentRecord:
    """
    Information about the experiment and the execution
    """

    id: int
    label: str
    script: ScriptViewer
    start_time: datetime
    end_time: Optional[datetime]
    story: str
    success: bool


@dataclass
class MetadataRecord:
    """
    A single record of metadata that was saved during the experiment
    """

    experiment_id: int
    id: str
    label: str
    stage: int
    data: Any
    time: datetime


@dataclass
class DebugRecord:
    """
    extra information about the experiment execution for better debugging
    """

    experiment_id: int
    id: int
    python_env: str
    python_history: str
    station_specs: str
    extra: str


@dataclass
class ResultRecord:
    """
    A single result that was saved during the experiment
    """

    experiment_id: int
    id: str
    label: str
    story: str
    stage: int
    data: Any
    time: datetime


@dataclass
class PlotRecord:
    """
    A single plot information and plotting instructions that was saved during the
    experiment
    """

    experiment_id: int
    id: int
    plot_data: Any = None
    generator: Optional[PlotGenerator] = None
    label: Optional[str] = None
    story: Optional[str] = None


@dataclass
class FigureRecord:
    """
    A single plotly figure that was saved during the experiment
    """

    experiment_id: int
    id: int
    figure: go.Figure
    time: datetime


@dataclass
class NodeResults:
    """
    All results of a specific node
    """

    execution_id: int
    results: Iterable[ResultRecord]


class DataReader(ABC):
    """
    An abstract class for Entropy database, defines the way entropy reads data.
    If a database implements both DataWriter and DataReader abstract classes,
    entropy will be able to read the results with a specific set of functions,
    and view them.
    """

    def __init__(self):
        super().__init__()

    def get_last_experiments(self, count: int, success: bool = None) -> DataFrame:
        """
            read n number of last experiments to a pandas dataframe
        :param count: number of experiments
        :param success: Optional filter for the success property.
        """
        return self.get_experiments_range(0, count, success)

    @abstractmethod
    def get_experiments_range(
        self, starting_from_index: int, count: int, success: bool = None
    ) -> DataFrame:
        """
            read a range of experiments to a pandas dataframe
        :param starting_from_index: experiment index to start from (sorted desc by time)
        :param count: number of experiments
        :param success: Optional filter for the success property.
        :return: A DataFrame containing one row per Experiment
        """
        pass

    @abstractmethod
    def get_experiment_record(self, experiment_id: int) -> Optional[ExperimentRecord]:
        """
            returns a single experiment record, or None if experiment not found
        :param experiment_id: the id of experiment
        """
        pass

    @abstractmethod
    def get_experiments(
        self,
        label: Optional[str] = None,
        start_after: Optional[datetime] = None,
        end_after: Optional[datetime] = None,
        success: Optional[bool] = None,
    ) -> Iterable[ExperimentRecord]:
        """
            get multiple experiments records according to any combination of parameters
            filters
        :param label: experiment label to filter by
        :param start_after: experiments start after specific time
        :param end_after: experiments ended after specific time
        :param success: experiment success criteria
        """
        pass

    @abstractmethod
    def get_results(
        self,
        experiment_id: Optional[int] = None,
        label: Optional[str] = None,
        stage: Optional[int] = None,
    ) -> Iterable[ResultRecord]:
        """
            get multiple results according to any combination of parameters filters

        :param experiment_id: results from specific experiment
        :param label: results label to filter by
        :param stage: results stage within the experiment
        """
        pass

    @abstractmethod
    def get_metadata_records(
        self,
        experiment_id: Optional[int] = None,
        label: Optional[str] = None,
        stage: Optional[int] = None,
    ) -> Iterable[MetadataRecord]:
        """
            get multiple metadata records according to any combination of parameters
            filters

        :param experiment_id: metadata from specific experiment
        :param label: metadata label to filter by
        :param stage: metadata stage within the experiment
        """
        pass

    @abstractmethod
    def get_last_result_of_experiment(
        self, experiment_id: int
    ) -> Optional[ResultRecord]:
        """
        returns the last (by time) results of the requested experiment
        """
        pass

    @abstractmethod
    def get_debug_record(self, experiment_id: int) -> Optional[DebugRecord]:
        """
        returns debug information about the requested experiment
        """
        pass

    # noinspection PyTypeChecker
    @abstractmethod
    def get_plots(self, experiment_id: int) -> List[PlotRecord]:
        """
        returns a list of all plots saved in the requested experiment
        """
        warn(
            "This method will soon be deprecated. Please use get_figures() instead",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        pass

    @abstractmethod
    def get_figures(self, experiment_id: int) -> List[FigureRecord]:
        """
        returns a list of all figures saved in the requested experiment
        """
        pass

    @abstractmethod
    def get_node_stage_ids_by_label(
        self, label: str, experiment_id: Optional[int] = None
    ) -> List[int]:
        """
            returns a list of the stage_id property values of all nodes with the
            given label
        :param label: results label
        :param experiment_id: optional to specify an experiment id, so
                only results from this experiment will be returned
        """
        pass

    def get_results_from_node(
        self,
        node_label: str,
        experiment_id: Optional[int] = None,
        result_label: Optional[str] = None,
    ) -> Iterable[NodeResults]:
        """
            returns multiple results with the given label, and combination of
            parameters filters

        :param node_label: the label of node
        :param experiment_id: optional to specify an experiment id, so
                only results from this experiment will be returned
        :param result_label: optional to specify a label of the results
        """
        node_stage_ids = self.get_node_stage_ids_by_label(node_label, experiment_id)
        if not node_stage_ids:
            raise KeyError(f"node {node_label} not found")
        nodes_results = []
        for stage_id in node_stage_ids:
            nodes_results.append(
                NodeResults(
                    stage_id,
                    self.get_results(
                        experiment_id,
                        label=result_label,
                        stage=stage_id,
                    ),
                )
            )
        return nodes_results


class ExperimentReader:
    """
    Reads results and data from a single experiment
    """

    def __init__(self, experiment_id: int, db: DataReader) -> None:
        """
            Reads results and data from a single experiment
        :param experiment_id: the id of experiment
        :param db: results database that implemented the DataReader abstract class
        """
        super().__init__()
        self._experiment_id = experiment_id
        self._data_reader: DataReader = db

    def get_experiment_info(self) -> ExperimentRecord:
        """
        returns a record of the experiment info
        """
        return self._data_reader.get_experiment_record(self._experiment_id)

    def get_debug_record(self) -> Optional[DebugRecord]:
        """
        returns a record of extra debugging information about the experiment
        """
        return self._data_reader.get_debug_record(self._experiment_id)

    def get_metadata_records(
        self, label: Optional[str] = None
    ) -> Iterable[MetadataRecord]:
        """
            returns an iterable of records with the given label
        :param label: metadata label
        """
        return self._data_reader.get_metadata_records(self._experiment_id, label)

    def get_results(self, label: Optional[str] = None) -> Iterable[ResultRecord]:
        """
            returns an iterable of results record with the given label
        :param label: result label
        """
        return self._data_reader.get_results(self._experiment_id, label)

    def get_plots(self) -> List[PlotRecord]:
        """
        returns a list of plot records that were saved for current experiment
        """
        warn(
            "This method will soon be deprecated. Please use get_plots() instead",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self._data_reader.get_plots(self._experiment_id)

    def get_figures(self) -> List[FigureRecord]:
        """
        returns a list of plotly figures that were saved for current experiment
        """
        return self._data_reader.get_figures(self._experiment_id)



import zmq
from sqlalchemy import create_engine

__all__ = [
    "zmq_context",
    "status",
    "node_name",
    "node_description",
    "node_dependancies",
    "entropy_identity",
    "playbook",
    "databucket",
    "data_server",
]

status = None
node_name = ""
node_description = ""
node_dependancies = []
node_icon = ""
entropy_identity = None
playbook = None
databucket = None
data_server = None

zmq_context_variable = None
runtime_data = None
executor_input = None  #: address of executor zmq communication channel
executor_output = None


def zmq_context():
    global zmq_context_variable
    if zmq_context_variable is None:
        zmq_context_variable = zmq.Context()
    return zmq_context_variable


def runtime_db_connect():
    global runtime_data
    SQLALCHEMY_DATABASE_URL = (
        f"postgresql://{data_server[0]}:{data_server[1]}"
        + f'@{data_server[2]}/{data_server[3]}?client_encoding="utf8"'
    )
    runtime_data = create_engine(SQLALCHEMY_DATABASE_URL).connect()


def runtime_db_close():
    runtime_data.close()



from entropylab.flame.inputs import Inputs
from entropylab.flame.workflow import Workflow

__all__ = ["BayesianEstimation"]


class BayesianEstimation(object):
    def __init__(self, workflow_node_unique_name, precorrected_data=None):
        """error correction algorithm

        :param precorrected_data: (JSON - STREAM) cirucuit description
        """
        self._command = "python3"
        self._bin = "errorCorrection.py"
        self._name = workflow_node_unique_name
        self._icon = "bootstrap/speedometer.svg"
        self._inputs = _Inputs(
            precorrected_data=precorrected_data,
        )
        self._outputs = _Outputs(self._name)
        self._host = {}
        Workflow._register_node(self)  # register the node in the workflow context

    def host(self, **kwargs):
        """Sets additional options for execution on the host."""
        for key, value in kwargs.items():
            self._host[key] = value
        return self

    @property
    def i(self):
        """Node inputs"""
        return self._inputs

    @property
    def o(self):
        """Node outputs"""
        return self._outputs


class _Inputs(object):
    def __init__(self, precorrected_data=None):
        self._inputs = Inputs()

        self._inputs.state(
            "precorrected_data", description="cirucuit description", units="JSON"
        )
        self._inputs.set(precorrected_data=precorrected_data)

    @property
    def precorrected_data(self):
        """Input: cirucuit description (JSON)"""
        return self._inputs.get("precorrected_data")

    @precorrected_data.setter
    def precorrected_data(self, value):
        """Input: cirucuit description (JSON)"""
        self._inputs.set(precorrected_data=value)


class _Outputs(object):
    def __init__(self, name):
        self._name = name
        self._outputs = [
            "correction_data",
        ]

    @property
    def correction_data(self):
        """Output: corrected data
        :return: (JSON)
        """
        return "#" + self._name + "/correction_data"



from __future__ import annotations

import abc
from typing import List, Dict
from entropylab.logger import logger

import pandas as pd

from entropylab import SqlAlchemyDB
from entropylab.pipeline.api.data_reader import PlotRecord, FigureRecord
from entropylab.dashboard.pages.results.auto_plot import auto_plot

MAX_EXPERIMENTS_NUM = 10000


class DashboardDataReader(abc.ABC):
    @abc.abstractmethod
    def get_last_experiments(
        self,
        max_num_of_experiments: int = MAX_EXPERIMENTS_NUM,
        success: bool = None,
    ) -> List[Dict]:
        pass

    @abc.abstractmethod
    def get_plot_and_figure_data(self, exp_id: int) -> List[PlotRecord]:
        pass


class SqlalchemyDashboardDataReader(DashboardDataReader):
    def __init__(self, connector: SqlAlchemyDB) -> None:
        super().__init__()
        self._db: SqlAlchemyDB = connector
        self._figures_cache = {}

    def get_last_experiments(
        self,
        max_num_of_experiments: int = MAX_EXPERIMENTS_NUM,
        success: bool = None,
    ) -> List[Dict]:
        experiments = self._db.get_last_experiments(max_num_of_experiments, success)
        experiments["success"] = experiments["success"].apply(
            lambda x: "✔️" if x else "❌"
        )
        experiments["start_time"] = pd.DatetimeIndex(
            experiments["start_time"]
        ).strftime("%Y-%m-%d %H:%M:%S")
        experiments["end_time"] = pd.DatetimeIndex(experiments["end_time"]).strftime(
            "%Y-%m-%d %H:%M:%S"
        )
        records = experiments.to_dict("records")
        return records

    def get_last_result_of_experiment(
        self,
        experiment_id,
    ):
        return self._db.get_last_result_of_experiment(experiment_id)

    def get_plot_and_figure_data(self, exp_id: int) -> List[PlotRecord | FigureRecord]:
        plots = self._db.get_plots(exp_id)
        if exp_id not in self._figures_cache:
            logger.debug(f"Figures cache miss. exp_id=[{exp_id}]")
            self._figures_cache[exp_id] = self._db.get_figures(exp_id)
        else:
            logger.debug(f"Figures cache hit. exp_id=[{exp_id}]")
        figures = self._figures_cache[exp_id]
        if len(plots) > 0 or len(figures) > 0:
            return [*plots, *figures]
        else:
            # TODO: auto_plot to produce figures, not plots
            last_result = self._db.get_last_result_of_experiment(exp_id)
            if last_result is not None and last_result.data is not None:
                plot = auto_plot(exp_id, last_result.data)
                return [plot]
            else:
                return []



import os.path
import pickle
from datetime import datetime
from enum import Enum
from typing import Optional, Any, Iterable, TypeVar, Callable, List

import h5py
import numpy as np

from entropylab import RawResultData
from entropylab.pipeline.api.data_reader import ResultRecord, MetadataRecord
from entropylab.pipeline.api.data_writer import Metadata
from entropylab.logger import logger
from entropylab.pipeline.results_backend.sqlalchemy.model import (
    ResultDataType,
    ResultTable,
    Base,
)

T = TypeVar("T", bound=Base)
R = TypeVar("R", ResultRecord, MetadataRecord)


def _experiment_from(dset: h5py.Dataset) -> int:
    return dset.attrs["experiment_id"]


def _id_from(dset: h5py.Dataset) -> str:
    return dset.name


def _stage_from(dset: h5py.Dataset) -> int:
    return dset.attrs["stage"]


def _label_from(dset: h5py.Dataset) -> str:
    return dset.attrs["label"]


def _story_from(dset: h5py.Dataset) -> str:
    if "story" in dset.attrs:
        return dset.attrs["story"]
    return ""


def _data_from(dset: h5py.Dataset) -> Any:
    data = dset[()]
    if dset.dtype.metadata and dset.dtype.metadata.get("vlen") == str:
        return dset.asstr()[()]
    elif dset.attrs.get("data_type") == ResultDataType.Pickled.value:
        return pickle.loads(data)
    elif dset.attrs.get("data_type") == ResultDataType.String.value:
        # un-picklable data is stored as HDF5 Opaque so turn to bytes then to string:
        return data.tobytes().decode(encoding="utf-8")
    else:
        return data


def _time_from(dset: h5py.Dataset) -> datetime:
    return datetime.fromisoformat(dset.attrs["time"])


def _build_result_record(dset: h5py.Dataset) -> ResultRecord:
    return ResultRecord(
        experiment_id=_experiment_from(dset),
        id=_id_from(dset),
        label=_label_from(dset),
        story=_story_from(dset),
        stage=_stage_from(dset),
        data=_data_from(dset),
        time=_time_from(dset),
    )


def _build_metadata_record(dset: h5py.Dataset) -> MetadataRecord:
    return MetadataRecord(
        experiment_id=_experiment_from(dset),
        id=_id_from(dset),
        label=_label_from(dset),
        stage=_stage_from(dset),
        data=_data_from(dset),
        time=_time_from(dset),
    )


def _get_all_or_single(group: h5py.Group, name: Optional[str] = None):
    """
    Returns all or one child from an h5py.Group

    Parameters
    ----------
    group group to get child or children from. Can be h5py.File itself.
    name name of child to get. If None, indicates all children should be retrieved.

    Returns
    -------
    A list of group children (either h5py.Group or h5py.Datasets)
    """
    if name is None:
        return list(group.values())
    else:
        if str(name) in group:
            return [group[str(name)]]
        else:
            return []


class EntityType(Enum):
    RESULT = 1
    METADATA = 2


class _HDF5Reader:
    def get_result_records(
        self,
        experiment_id: Optional[int] = None,
        stage: Optional[int] = None,
        label: Optional[str] = None,
    ) -> Iterable[ResultRecord]:
        return self._get_records(
            EntityType.RESULT, _build_result_record, experiment_id, stage, label
        )

    def get_metadata_records(
        self,
        experiment_id: Optional[int] = None,
        stage: Optional[int] = None,
        label: Optional[str] = None,
    ) -> Iterable[MetadataRecord]:
        return self._get_records(
            EntityType.METADATA, _build_metadata_record, experiment_id, stage, label
        )

    def _get_records(
        self,
        entity_type: EntityType,
        record_build_func: Callable,
        experiment_id: Optional[int] = None,
        stage: Optional[int] = None,
        label: Optional[str] = None,
    ) -> Iterable[T]:
        entities = []
        if experiment_id:
            experiment_ids = [experiment_id]
        else:
            experiment_ids = self._list_experiment_ids_in_fs()
        for experiment_id in experiment_ids:
            entities += self._get_experiment_entities(
                entity_type, record_build_func, experiment_id, stage, label
            )
        return sorted(entities, key=lambda entity: entity.experiment_id)

    def _list_experiment_ids_in_fs(self) -> List[int]:
        # noinspection PyUnresolvedReferences
        dir_list = os.listdir(self._path)
        # TODO: Better validation of experiment ids
        exp_files = filter(lambda f: f.endswith(".hdf5"), dir_list)
        experiment_ids = list(map(lambda f: f[:-5], exp_files))
        return experiment_ids

    def _get_experiment_entities(
        self,
        entity_type: EntityType,
        convert_from_dset: Callable,
        experiment_id: int,
        stage: Optional[int] = None,
        label: Optional[str] = None,
    ) -> Iterable[T]:
        dsets = []
        try:
            # noinspection PyUnresolvedReferences
            file = self._open_hdf5(experiment_id, "r")
        except FileNotFoundError:
            logger.error(f"HDF5 file for experiment_id [{experiment_id}] was not found")
        else:
            with file:
                stage_groups = _get_all_or_single(file, stage)
                for stage_group in stage_groups:
                    label_groups = _get_all_or_single(stage_group, label)
                    for label_group in label_groups:
                        dset_name = entity_type.name.lower()
                        dset = label_group[dset_name]
                        dsets.append(convert_from_dset(dset))
        return dsets

    def get_last_result_of_experiment(
        self, experiment_id: int
    ) -> Optional[ResultRecord]:
        results = list(self.get_result_records(experiment_id, None, None))
        if results and len(results) > 0:
            results.sort(key=lambda x: x.time, reverse=True)
            return results[0]
        else:
            return None


class _HDF5Writer:
    def save_result(self, experiment_id: int, result: RawResultData) -> str:
        # noinspection PyUnresolvedReferences
        with self._open_hdf5(experiment_id, "a") as file:
            return self._save_entity_to_file(
                file,
                EntityType.RESULT,
                experiment_id,
                result.stage,
                result.label,
                result.data,
                datetime.now(),
                result.story,
            )

    def save_metadata(self, experiment_id: int, metadata: Metadata):
        # noinspection PyUnresolvedReferences
        with self._open_hdf5(experiment_id, "a") as file:
            return self._save_entity_to_file(
                file,
                EntityType.METADATA,
                experiment_id,
                metadata.stage,
                metadata.label,
                metadata.data,
                datetime.now(),
            )

    def _save_entity_to_file(
        self,
        file: h5py.File,
        entity_type: EntityType,
        experiment_id: int,
        stage: int,
        label: str,
        data: Any,
        time: datetime,
        story: Optional[str] = None,
        migrated_id: Optional[str] = None,
    ) -> str:
        path = f"/{stage}/{label}"
        label_group = file.require_group(path)
        dset = self._create_dataset(label_group, entity_type, data)
        dset.attrs.create("experiment_id", experiment_id)
        dset.attrs.create("stage", stage)
        dset.attrs.create("label", label)
        dset.attrs.create("time", time.astimezone().isoformat())
        if story:
            dset.attrs.create("story", story or "")
        if migrated_id:
            dset.attrs.create("migrated_id", migrated_id or "")
        return dset.name

    def _create_dataset(
        self, group: h5py.Group, entity_type: EntityType, data: Any
    ) -> h5py.Dataset:
        name = entity_type.name.lower()
        try:
            dset = group.create_dataset(name=name, data=data)
        except TypeError:
            data_type, pickled = self._pickle_data(data)
            # np.void turns our string to bytes (HDF5 Opaque):
            dset = group.create_dataset(name=name, data=np.void(pickled))
            dset.attrs.create("data_type", data_type.value, dtype="i2")
        return dset

    @staticmethod
    def _pickle_data(data: Any) -> (bytes, ResultDataType):
        try:
            pickled = pickle.dumps(data)
            data_type = ResultDataType.Pickled
        except Exception as ex:
            logger.debug("Could not pickle data, defaulting to __repr__()", ex)
            pickled = data.__repr__().encode(encoding="utf-8")
            data_type = ResultDataType.String
        return data_type, pickled


class _HDF5Migrator(_HDF5Writer):
    def migrate_result_rows(self, rows: Iterable[ResultTable]) -> None:
        self.migrate_rows(EntityType.RESULT, rows)

    def migrate_metadata_rows(self, rows: Iterable[Metadata]) -> None:
        self.migrate_rows(EntityType.METADATA, rows)

    def migrate_rows(self, entity_type: EntityType, rows: Iterable[T]) -> None:
        if rows is not None and len(list(rows)) > 0:
            for row in rows:
                if not row.saved_in_hdf5:
                    record = row.to_record()
                    # noinspection PyUnresolvedReferences
                    with self._open_hdf5(record.experiment_id, "a") as file:
                        hdf5_id = self._migrate_record(file, entity_type, record)
                    logger.debug(
                        f"Migrated ${entity_type.name} with id [{row.id}] "
                        f"to HDF5 with id [{hdf5_id}]"
                    )

    def _migrate_record(
        self, file: h5py.File, entity_type: EntityType, record: R
    ) -> str:
        return self._save_entity_to_file(
            file,
            entity_type,
            record.experiment_id,
            record.stage,
            record.label,
            record.data,
            record.time,
            getattr(record, "story", None),
            record.id,
        )

    def _migrate_result_record(
        self, file: h5py.File, result_record: ResultRecord
    ) -> str:
        return self._save_entity_to_file(
            file,
            EntityType.RESULT,
            result_record.experiment_id,
            result_record.stage,
            result_record.label,
            result_record.data,
            result_record.time,
            result_record.story,
            result_record.id,
        )

    def migrate_from_per_project_hdf5_to_per_experiment_hdf5_files(
        self, old_global_hdf5_file_path
    ):
        logger.debug(
            f"Migrating global .hdf5 file {old_global_hdf5_file_path} "
            "to per-experiment .hdf5 files"
        )
        with h5py.File(old_global_hdf5_file_path, "r") as file:
            if "experiments" in file:
                top_group = file["experiments"]
                for exp_group in top_group.values():
                    experiment_id = int(exp_group.name[13:])
                    f"Migrating results and metadata for experiment id {experiment_id}"
                    # noinspection PyUnresolvedReferences
                    with self._open_hdf5(experiment_id, "a") as exp_file:
                        for stage_group in exp_group.values():
                            exp_group.copy(stage_group, exp_file)
        new_filename = f"{old_global_hdf5_file_path}.bak"
        logger.debug(f"Renaming global .hdf5 file to [{new_filename}]")
        os.rename(old_global_hdf5_file_path, new_filename)
        logger.debug("Global .hdf5 file migration done")


class HDF5Storage(_HDF5Reader, _HDF5Migrator, _HDF5Writer):
    def __init__(self, path=None):
        """Initializes a new storage class instance  for storing experiment results
                 and metadata in HDF5 files.

        :param path: filesystem path to a directory where HDF5 files reside. If no path
                 is given or the path is empty, HDF5 files are stored in memory only.
        """
        if path is None or path == "":  # memory files
            self._path = "./entropy_temp_hdf5"
            self._in_memory_mode = True
        else:  # filesystem
            self._path = path
            os.makedirs(self._path, exist_ok=True)
            self._in_memory_mode = False

    def _open_hdf5(self, experiment_id: int, mode: str) -> h5py.File:
        path = self._build_hdf5_filepath(experiment_id)
        try:
            if self._in_memory_mode:
                """Note that because backing_store=False, self._path is ignored & no
                file is saved on disk.
                See https://docs.h5py.org/en/stable/high/file.html#file-drivers
                """
                return h5py.File(path, mode, driver="core", backing_store=False)
            else:
                return h5py.File(path, mode)
        except FileNotFoundError:
            logger.exception(f"HDF5 file not found at '{path}'")
            raise

    def _build_hdf5_filepath(self, experiment_id: int) -> str:
        return os.path.join(self._path, f"{experiment_id}.hdf5")



import entropynodes.library as expNodes
from entropylab.flame.workflow import Workflow

wf = Workflow("A day in office", description="What happens in local office")

boss = expNodes.GrumpyAdministrator("boss")
clerk = expNodes.CheerfulNode("clerk", customer=boss.o.clerk_request)
boss.i.clerk_salary = clerk.o.requested_salary

wf.register()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from difflib import ndiff
from typing import Any, List, Optional

import jsonpickle
import jsonpickle.ext.numpy as jsonpickle_numpy
import jsonpickle.ext.pandas as jsonpickle_pandas

jsonpickle_numpy.register_handlers()
jsonpickle_pandas.register_handlers()

Entropy_Resource_Name = "entropy_name"


@dataclass
class Parameter:
    """"""

    name: str
    unit: Optional[str]
    step: float
    scale: float
    offset: float
    vals: Any
    submodule: str = None


@dataclass
class Function:
    """"""

    name: str
    parameters = None


@dataclass
class DriverSpec:
    """"""

    parameters: List[Parameter]
    functions: List[Function]
    undeclared_functions: List[Function]


class Resource(ABC):
    """
    An abstract class with extra functionality.
    Entropy will be able to use the snapshot and revert_to_snapshot in experiments,
    making it easier to control the resource, and save metadata
    The he resource might be  an actual instrument in the lab as well.
    Entropy will call the "connect" when starting the driver,
    and the "teardown" on close, so no connections will be left open
    """

    def __init__(self, **kwargs):
        super().__init__()
        self._entropy_name = kwargs.get(Entropy_Resource_Name, "")

    @abstractmethod
    def connect(self):
        """
        Start the driver, open needed connections or call any other functions that
        are needed for using the driver
        """
        pass

    @abstractmethod
    def teardown(self):
        """
        Gracefully close all driver connections and links
        """
        pass

    def get_dynamic_driver_specs(self) -> DriverSpec:
        """
        add metadata for the driver functionality, if discovered dynamically
        """
        pass

    def __del__(self):
        self.teardown()

    def get_instance(self):
        """
            Returns the actual instance of the driver.

            If the driver is wrapping a different instrument, should return the
            actual instance.

        :return:
        """
        return self

    @abstractmethod
    def snapshot(self, update: bool) -> str:
        """
            serialize the current resource state to string
            The completeness of the state is up to the implementation.
            Best practice will be to create the most accurate representation,
            allowing "revert_to_snapshot" function to revert the resource to the
            same exact state
        :param update: Whether to update the resource state actively, or use the
                        cached data
        """
        pass

    def revert_to_snapshot(self, snapshot: str):
        """
            revert the current resource to a different state, using a saved
            snapshot
        :param snapshot: a serialized state of the resource, created by
                        the "snapshot" function
        """
        raise NotImplementedError(
            f"resource {self.__class__.__qualname__} has not implemented revert to snapshot"
        )

    def diff_from_snapshot(self, other_snapshot: str):
        """
            calculates the difference between the given snapshot and current resource state
        :param other_snapshot: a snapshot of this resource
        """
        snapshot = self.snapshot(False)
        return ndiff(snapshot, other_snapshot)

    def set_entropy_name(self, name: str):
        """
        helper function for entropy, to set names of resources
        """
        self._entropy_name = name


class PickledResource(Resource):
    """
    An implemetation of "snapshot" function using jsonpickle
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def snapshot(self, update: bool = True) -> str:
        """
            serialize the current resource state to string with jsonpickle
             package.
        :param update: Whether to update the resource state actively, or use the
                        cached data
        """
        return jsonpickle.encode(self)

    @abstractmethod
    def connect(self):
        """
        Start the driver, open needed connections or call any other functions that
        are needed for using the driver
        """
        pass

    @abstractmethod
    def teardown(self):
        """
        Gracefully close all driver connections and links
        """
        pass

    def revert_to_snapshot(self, snapshot: str):
        """
            revert the current resource to a different state, using a saved
            snapshot
        :param snapshot: a serialized state of the resource, created by
                        the "snapshot" function
        """
        raise NotImplementedError(
            f"resource {self.__class__.__qualname__} has not implemented revert to snapshot"
        )



"""
Definitions of enumerated types used on several places
"""

from enum import Enum


class InputType(Enum):
    STATE = 1
    STREAM = 2


class ioFormat(Enum):
    VALUES = 1
    JUST_STRUCTURE = 2



import entropynodes.library as expNodes
from entropylab.flame.workflow import Workflow

wf = Workflow(
    "Scan with error correction",
    description="Performs Bayesian estimation for the optimal parameters "
    "and uses this during the circuit run.",
)

c = expNodes.CalibrateSystem("calibration_1")

scan_list = expNodes.TaskList(
    "scan_list",
    calibration=c.o.calibration_status,
)

laser = expNodes.ExternalScan(
    "external_instruments", set_point=scan_list.o.scan_setpoint
)
scan_list.i.laser_setpoint_locked = laser.o.setpoint_reached

qpu = expNodes.QPUcircuitRunner("QPU", circuit_param=scan_list.o.circuit_specification)
scan_list.i.circuit_done = qpu.o.circuit_finished

errCor = expNodes.BayesianEstimation("correction_1")
errCor.i.precorrected_data = qpu.o.precorrected_data
qpu.i.error_correction = errCor.o.correction_data

final_report = expNodes.FinalDataAnalysis("final_report", data=qpu.o.final_data)

wf.register()



from .enum_types import InputType, ioFormat
from . import nodeio_context
import json
import zmq
import msgpack
import time

__all__ = ["Inputs"]


class Inputs:

    __local_instances = []  # where to get values from when used in stand-alone mode
    __external_intances = []  # where to get values when used as a module

    @classmethod
    def _register(cls, instance):
        cls.__local_instances.append(instance)
        return len(cls.__local_instances) - 1

    @classmethod
    def _schema(cls):
        inputList = []
        for v in cls.__local_instances:
            # TODO: handle properly Inputs instances in Inputs lists
            typeDict = {}
            for k, e in v.input_type.items():
                typeDict[k] = e.value
            inputList.append(
                {"description": v.description, "units": v.units, "type": typeDict}
            )
        return inputList

    @classmethod
    def _set_as_external_input(cls, instance):
        cls.__local_instances.remove(instance)
        cls.__external_intances.append(instance)
        instance.__index = len(cls.__external_intances) - 1

    @classmethod
    def _connections(cls):
        return cls.__local_instances[0].connections

    def __init__(self):
        self.__index = self._register(self)
        if self.__index < len(self.__external_intances):
            # overwrite inputs with externally imposed inputs
            self.values = self.__external_intances[self.__index].values
            self.units = self.__external_intances[self.__index].units
            self.value_set = self.__external_intances[self.__index].value_set
            self.description = self.__external_intances[self.__index].description
            self.input_type = self.__external_intances[self.__index].input_type
        else:
            # use locally defined inputs
            self.values = {}
            self.units = {}
            self.value_set = {}
            self.description = {}
            self.input_type = {}
        self.connections = {}  # communication middle ware providing inputs
        self.convert_to = {}  # data type conversion from plain text
        self.last_change = {}  # since the last reading, or None if no new updates

    def state(self, name: str, description="", units=""):
        """Input that holds only last received value. It is not consumed on
        reading. It doesn't block if it has been set in the past. Useful for
        stateful variables.
        """
        return self.__input(name, InputType.STATE, description=description, units=units)

    def stream(self, name: str, description="", units=""):
        """Input that accumulates received values for consumption (by get method),
        It blocks on read if until there are elements to be consumed. Useful for
        data streams and events.
        """
        return self.__input(
            name, InputType.STREAM, description=description, units=units
        )

    def __input(self, name: str, input_type: InputType, description="", units=""):
        """Returns a function whose value can be obtained later to obtain
        input value.
        """

        if name.find(" ") != -1 or name.find("/") != -1:
            raise ValueError(
                "Input name cannot contain spaces or / characters. "
                "Use underscore_names or camelCase."
            )
        if self.__index >= len(self.__external_intances):
            # if we don't have external inputs, initialize values
            if name in self.values:
                raise ValueError(f"Input varible {name} has been alredy defined.")
            self.values[name] = None
            self.value_set[name] = False
        else:
            # we have external inputs
            if (
                type(self.values[name][0]) == str
                and len(self.values[name][0]) > 0
                and self.values[name][0][0] == "#"
            ):
                # if they are runtime variables, initialize connections
                self.value_set[name] = False
                context = nodeio_context.zmq_context()
                socket = context.socket(zmq.SUB)
                if input_type == InputType.STATE:
                    # state variables should get only the last, up-to date
                    # variable
                    socket.setsockopt(zmq.CONFLATE, 1)
                socket.setsockopt(zmq.LINGER, 0)
                socket.connect(self.values[name][0][1:])
                socket.subscribe("")
                self.connections[name] = socket

        self.description[name] = description
        self.units[name] = units
        self.input_type[name] = input_type
        return lambda: self.get(name)

    def get(self, name: str):
        """Returns value of the input

        For state variables, returns currently value, blocks only if it
        has not been still resolved at all (first resolution).

        For stream variables, returns current value, blocking if no value is
        available until value is received. If module is used in stand-alone
        mode, instead of blocking, once the input values are all used,
        it will throw error.
        """
        if self.__index >= len(self.__external_intances):
            assert self.value_set[name], f"Inputs '{name}' has not been defined yet."
            if len(self.values[name]) == 0:
                print(
                    "\tENTROPYLAB - INPUTS: End of the input queue "
                    f"{nodeio_context.node_name}/{name} is reached, "
                    "terminating test run of the node."
                )
                exit()
            if self.input_type[name] == InputType.STREAM:
                print(f"\tENTROPYLAB - STREAM INPUT {nodeio_context.node_name}/{name}")
            else:
                print(f"\tENTROPYLAB - STATE INPUT {nodeio_context.node_name}/{name}")
            if len(self.values[name]) == 1 and self.input_type[name] == InputType.STATE:
                # InputType.STATE
                # we don't have sequence of states to test
                # assume that last state stays from now on
                value = self.values[name][0]
            else:
                value = self.values[name].pop(0)

            print(f"\t{value}")
            return value

        else:
            # node is part of workflow
            if name in self.connections:
                # requested input is runtime variable
                if self.input_type[name] == InputType.STREAM:
                    if self.value_set[name]:
                        # there is cached value
                        self.value_set[name] = False
                        return self.values[name]

                    # else: block until input is resolved
                    try:
                        self.values[name] = self.convert_to[name](
                            msgpack.unpackb(self.connections[name].recv())
                        )
                    except Exception as e:
                        print(repr(e))
                        exit(1)
                    return self.values[name]
                elif self.input_type[name] == InputType.STATE:
                    if self.value_set[name]:
                        # we already have some state runtime variable
                        # just check in non-blocking manner if there is update
                        # to this value
                        try:
                            # check for updates
                            self.values[name] = self.convert_to[name](
                                msgpack.unpackb(
                                    self.connections[name].recv(flags=zmq.NOBLOCK)
                                )
                            )
                        except zmq.error.ZMQError as e:
                            if e.errno == zmq.EAGAIN:
                                # the state value has not been updated since last
                                # get call
                                self.last_change[name] = None
                                return self.values[name]
                            else:
                                raise Exception(repr(e))
                        self.last_change[name] = None
                        return self.values[name]
                    else:
                        # this is first resolution of the state runtime variable
                        # block here until the value is received
                        try:
                            self.values[name] = self.convert_to[name](
                                msgpack.unpackb(self.connections[name].recv())
                            )
                        except Exception as e:
                            print(repr(e))
                            exit(1)
                        self.last_change[name] = None
                        self.value_set[name] = True
                        return self.values[name]

            # requested input is parameter
            return self.convert_to[name](self.values[name][0])

    def set(self, **kwargs):
        """Sets the value of the input.

        State variables will save only the latest set value.
        Stream varialbes will save all the variables in a queue and make them
        available for consummers.
        """
        for key, value in kwargs.items():
            if key not in self.values:
                raise ValueError(f"The input {key} has not been defined.")
            # deduce type for conversion
            if value is not None:
                self.convert_to[key] = type(value)
            else:
                self.values[key] = None
                self.value_set[key] = False
                return
            # if we are using external instance of inputs prevent overwritting of the parameters
            if self.__index >= len(self.__external_intances):
                if self.values[key] is None:
                    self.values[key] = []

                self.values[key].append(value)

                self.value_set[key] = True
        return

    def updated(self, name):
        """Checks if value has been updated.

        Note that this does not check the variable value, just if the variable
        has been set again since the last time we get it.

        :returns: bool
        """
        if self.input_type[name] == InputType.STATE:
            if name not in self.connections:
                # this is either a parameter (not a runtime variable)
                return False
            # else this is a run-time variable that has not been yet resolved
            if name in self.last_change and self.last_change[name] is not None:
                return True
            # check for updates
            try:
                self.values[name] = self.convert_to[name](
                    msgpack.unpackb(self.connections[name].recv(flags=zmq.NOBLOCK))
                )
                self.last_change[name] = time.time()
            except zmq.error.ZMQError as e:
                if e.errno == zmq.EAGAIN:
                    # the state value has not been updated since last
                    # get call
                    return False
                else:
                    raise Exception(repr(e))
            return True
        else:
            # case when self.input_type[name] == InputType.STREAM
            if self.value_set[name]:
                # there is unconsumed stream value
                return True
            else:
                # check if there is update
                try:
                    self.values[name] = self.convert_to[name](
                        msgpack.unpackb(self.connections[name].recv(flags=zmq.NOBLOCK))
                    )
                    self.value_set[name] = True
                    return True
                except zmq.error.ZMQError as e:
                    if e.errno == zmq.EAGAIN:
                        # this value has not been set yet
                        return False
                    else:
                        raise Exception(repr(e))

    def remove(self, name):
        """Removes named input

        Values of this input will be deleted and this named input will be
        removed from list of available inputs.
        Note that any get calls to this name will be undefined, unless input
        with this name is added at later time.
        """
        if name in self.values:
            del self.values[name]
            del self.value_set[name]
            del self.description[name]
            del self.units[name]
            del self.input_type[name]

    def reset(self, name):
        """Resets input value to not set, keeping the input variable"""
        if name in self.values:
            self.values[name] = None
            self.value_set[name] = False

    def reset_all(self):
        """Resets all set input values to not set, keeping the input variable"""
        for name, _value in self.values.items():
            self.values[name] = None
            self.value_set[name] = False

    def defines(self, name):
        """Returns true if input exists and it is defined."""
        return (name in self.values) and (self.value_set[name])

    def completely_defined(self):
        return self.get_as_JSON("undefined") == {}

    def print_all(self, indent=2, sort_keys=True):
        inputs = self.get_as_JSON("all")
        return json.dumps(inputs, indent=indent, sort_keys=sort_keys)

    def print_defined(self, indent=2, sort_keys=True):
        inputs = self.get_as_JSON("defined")
        return json.dumps(inputs, indent=indent, sort_keys=sort_keys)

    def print_undefined(self, indent=2, sort_keys=True):
        inputs = self.get_as_JSON("undefined")
        return json.dumps(inputs, indent=indent, sort_keys=sort_keys)

    def get_as_JSON(self, what="all", format=ioFormat.VALUES):
        def _filter_all(value_set):
            return True

        def _filter_defined(value_set):
            return value_set

        def _filter_undefined(value_set):
            return not value_set

        if what == "all":
            filter = _filter_all
        elif what == "defined":
            filter = _filter_defined
        elif what == "undefined":
            filter = _filter_undefined
        else:
            raise ValueError(
                "Only all, defined and undefined are valid inputs " "for get_as_JSON"
            )

        outJSON = {}
        for k, v in self.values.items():
            if self.value_set[k]:
                v = v[0]  # only current value if we have event sequence
            if filter(self.value_set[k]):
                if isinstance(v, Inputs):
                    outJSON[k] = v.get_as_JSON(what=what, format=format)
                elif isinstance(v, list):
                    outJSON[k] = []
                    for element in v:
                        if isinstance(element, Inputs):
                            outJSON[k].append(
                                element.get_as_JSON(what=what, format=format),
                            )
                        else:
                            outJSON[k].append(element)
                else:
                    outJSON[k] = v
        return outJSON

    def load_JSON(self):
        pass

    def view_as_HTML(self, enrich=True):
        """If enrich is True, add docstrings"""
        pass

    def print_docstring(self, name):

        pass

    def __repr__(self) -> str:
        return "Inputs"

    def print_instances(self):
        print(self.__local_instances)



import os
import shutil

print("\n = = = = = = entropylab empty project = = = = =")
print("Writting simple example project in the current directory...")

origin = os.path.join(
    os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "empty"
)
destination = os.getcwd()

shutil.copy(os.path.join(origin, "node.py"), destination)
shutil.copy(os.path.join(origin, "workflow.py"), destination)

print("\t1. First write and execute\tpython node.py\t\tto test single node")
print("\t2. Then execute\t\t\tpython workflow.py\tto generate list of parameters")
print("\t3. Write any unresolved input in parameters.json  ")
print("\t4. Execute workflow with\tpython -m entropylab.flame.execute ")
print("\t  (to interrupt and terminate execution press Ctrl+C)")
print("\n")



import logging


class _Config:
    port_number = 9000
    node_status_dict = {}
    execution_active = True
    job_id = None
    DATABASE_NAME = None
    SQLALCHEMY_DATABASE_URL = None


logger = logging.getLogger("ExecuteLogger")



import os
import shutil

print("\n = = = = = = entropylab Python NodeIO = = = = = =")
print("Writting empty node.py example in the current directory...")

origin = os.path.join(
    os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "empty"
)
destination = os.getcwd()

shutil.copy(os.path.join(origin, "node.py"), destination)

print("\tdone!")
print("\n")



from entropylab.flame.inputs import Inputs
from entropylab.flame.workflow import Workflow

__all__ = ["TaskList"]


class TaskList(object):
    def __init__(
        self,
        workflow_node_unique_name,
        calibration=None,
        scan_points=None,
        laser_setpoint_locked=None,
        circuit_done=None,
    ):
        """Does scan of parameters and triggers circuit execution

        :param calibration: (status - STATE) node will execute only if it receives positive calibration status
        :param scan_points: (MHz - STATE) an array for setpoint execution
        :param laser_setpoint_locked: (bool - STREAM) Triger when laser setpoint is set
        :param circuit_done: (bool - STREAM) Triger when circuit execution is done
        """
        self._command = "python3"
        self._bin = "sequence.py"
        self._name = workflow_node_unique_name
        self._icon = "bootstrap/list-task.svg"
        self._inputs = _Inputs(
            calibration=calibration,
            scan_points=scan_points,
            laser_setpoint_locked=laser_setpoint_locked,
            circuit_done=circuit_done,
        )
        self._outputs = _Outputs(self._name)
        self._host = {}
        Workflow._register_node(self)  # register the node in the workflow context

    def host(self, **kwargs):
        """Sets additional options for execution on the host."""
        for key, value in kwargs.items():
            self._host[key] = value
        return self

    @property
    def i(self):
        """Node inputs"""
        return self._inputs

    @property
    def o(self):
        """Node outputs"""
        return self._outputs


class _Inputs(object):
    def __init__(
        self,
        calibration=None,
        scan_points=None,
        laser_setpoint_locked=None,
        circuit_done=None,
    ):
        self._inputs = Inputs()

        self._inputs.state(
            "calibration",
            description="node will execute only if it receives positive calibration status",
            units="status",
        )
        self._inputs.set(calibration=calibration)

        self._inputs.state(
            "scan_points", description="an array for setpoint execution", units="MHz"
        )
        self._inputs.set(scan_points=scan_points)

        self._inputs.state(
            "laser_setpoint_locked",
            description="Triger when laser setpoint is set",
            units="bool",
        )
        self._inputs.set(laser_setpoint_locked=laser_setpoint_locked)

        self._inputs.state(
            "circuit_done",
            description="Triger when circuit execution is done",
            units="bool",
        )
        self._inputs.set(circuit_done=circuit_done)

    @property
    def calibration(self):
        """Input: node will execute only if it receives positive calibration status (status)"""
        return self._inputs.get("calibration")

    @calibration.setter
    def calibration(self, value):
        """Input: node will execute only if it receives positive calibration status (status)"""
        self._inputs.set(calibration=value)

    @property
    def scan_points(self):
        """Input: an array for setpoint execution (MHz)"""
        return self._inputs.get("scan_points")

    @scan_points.setter
    def scan_points(self, value):
        """Input: an array for setpoint execution (MHz)"""
        self._inputs.set(scan_points=value)

    @property
    def laser_setpoint_locked(self):
        """Input: Triger when laser setpoint is set (bool)"""
        return self._inputs.get("laser_setpoint_locked")

    @laser_setpoint_locked.setter
    def laser_setpoint_locked(self, value):
        """Input: Triger when laser setpoint is set (bool)"""
        self._inputs.set(laser_setpoint_locked=value)

    @property
    def circuit_done(self):
        """Input: Triger when circuit execution is done (bool)"""
        return self._inputs.get("circuit_done")

    @circuit_done.setter
    def circuit_done(self, value):
        """Input: Triger when circuit execution is done (bool)"""
        self._inputs.set(circuit_done=value)


class _Outputs(object):
    def __init__(self, name):
        self._name = name
        self._outputs = [
            "circuit_specification",
            "scan_setpoint",
            "status",
        ]

    @property
    def circuit_specification(self):
        """Output: Triggers circuit execution by providing information for run
        :return: (JSON)
        """
        return "#" + self._name + "/circuit_specification"

    @property
    def scan_setpoint(self):
        """Output: Laser setpoint
        :return: (MHz)
        """
        return "#" + self._name + "/scan_setpoint"

    @property
    def status(self):
        """Output: window into what node is doing
        :return: (json)
        """
        return "#" + self._name + "/status"



import abc
from itertools import count
from typing import Any

from plotly import graph_objects as go

from entropylab.pipeline.api.data_writer import (
    DataWriter,
    RawResultData,
    Metadata,
    PlotSpec,
)
from entropylab.components.lab_topology import ExperimentResources


class EntropyContext:
    """
    An interface for the environment of current execution.
    Using the context, user can save data to the database
        and use resources of the experiments.
    """

    def __init__(
        self,
        exp_id: int,
        data_writer: DataWriter,
        experiment_resources: ExperimentResources,
        stage_id: int,
        context_factory,
    ) -> None:
        """
            An interface for the environment of current execution.
        :param exp_id: experiment id
        :param data_writer: data writer used for this experiment
        :param experiment_resources: the resources used in this experiment
        """
        super().__init__()
        self._data_writer = data_writer
        self._exp_id = exp_id
        self._experiment_resources = experiment_resources
        self._stage_id = stage_id
        self._context_factory = context_factory

    def add_result(self, label: str, data: Any, story: str = None):
        """
        saves a new result from this experiment in the database
        :param label: result label
        :param data: result data
        :param story: story about the result
        """
        self._data_writer.save_result(
            self._exp_id, RawResultData(label, data, self._stage_id, story)
        )

    def add_metadata(self, label: str, metadata: Any):
        """
        saves a new metadata from this experiment in the database
        """
        self._data_writer.save_metadata(
            self._exp_id, Metadata(label, self._stage_id, metadata)
        )

    def add_plot(self, plot: PlotSpec, data: Any):
        """
            saves a new plot from this experiment in the database

        :param plot: description and plotting instructions
        :param data: the data for plotting
        """
        self._data_writer.save_plot(self._exp_id, plot, data)

    def add_figure(self, figure: go.Figure) -> None:
        """
            saves a new figure from this experiment in the database

        :param figure: a Plotly Figure object to be saved
        """
        self._data_writer.save_figure(self._exp_id, figure)

    def get_resource(self, name):
        """
            Return an experiment resource (shared lab resource or execution temp resource).
            User can use this resource within the experiment
        :param name: resource name
        """
        return self._experiment_resources.get_resource(name)

    def has_resource(self, name) -> bool:
        """
            True if the resource exist and was added to the current experiment
        :param name: resource name
        """
        return self._experiment_resources.has_resource(name)

    def _get_stage_id(self):
        return self._stage_id


class _EntropyContextFactory:
    def __init__(
        self,
        exp_id: int,
        data_writer: DataWriter,
        experiment_resources: ExperimentResources,
    ) -> None:
        super().__init__()
        self._data_writer = data_writer
        self._exp_id = exp_id
        self._experiment_resources = experiment_resources
        self._stage_iter = count(start=0, step=1)

    def create(self) -> EntropyContext:
        return EntropyContext(
            self._exp_id,
            self._data_writer,
            self._experiment_resources,
            next(self._stage_iter),
            self,
        )


class ExperimentExecutor(abc.ABC):
    """
    Abstract class for executor,
    Every experiment definition should include an executor.
    on every execution, a new executor will be created.
    """

    def __init__(
        self,
    ) -> None:
        super().__init__()

    @abc.abstractmethod
    def execute(self, entropy_context_factory: _EntropyContextFactory) -> Any:
        """
            actual execution of the experiment
        :param entropy_context_factory: the context of current execution
        """
        pass

    @property
    @abc.abstractmethod
    def failed(self) -> bool:
        """
        :return: True if this execution failed
        """
        pass



import abc
import json
from datetime import datetime
from typing import Optional

from entropylab.pipeline.api.data_reader import DataReader
from entropylab.pipeline.api.data_writer import (
    DataWriter,
    ExperimentInitialData,
    RawResultData,
    ExperimentEndData,
)
from entropylab.pipeline.api.errors import EntropyError
from entropylab.pipeline.api.execution import ExperimentExecutor, _EntropyContextFactory
from entropylab.pipeline.api.memory_reader_writer import MemoryOnlyDataReaderWriter
from entropylab.components.lab_topology import ExperimentResources
from entropylab.logger import logger


class _Experiment:
    """
    Instance of the experiment, executed according to the definition.
    """

    def __init__(self, definition: "ExperimentDefinition", db: DataWriter):
        """
        Instance of the experiment, executed according to the definition.

        :param definition: the experiment definition that includes all relevant
                        information for executing
        :param db: db implementation to save results to
        """
        self._start_time: Optional[datetime] = None
        self._end_time: Optional[datetime] = None
        if not isinstance(db, DataWriter):
            raise TypeError(f"db must be of type {DataWriter}")
        self._definition: ExperimentDefinition = definition
        self._data_writer: DataWriter = db
        self._experiment_resources: ExperimentResources = (
            definition.get_experiment_resources()
        )
        if definition._user is None:
            self._user = ""
        else:
            self._user: str = definition._user
        self._id: int = -1
        self._executor = self._definition._get_execution_instructions()

    def run(self) -> bool:
        """
            runs the current experiment
        :return: success
        """
        if self._start_time is not None:
            raise EntropyError("Can not run the same experiment twice")
        try:
            self._start_time = datetime.now()
            initial_data = ExperimentInitialData(
                label=self._definition.label,
                user=self._user,
                lab_topology=json.dumps(
                    self._experiment_resources._serialize_resources_snapshot()
                ),
                script=self._definition.serialize(),
                start_time=self._start_time,
                story=self._definition.story,
            )
            self._id = self._data_writer.save_experiment_initial_data(initial_data)

            self._experiment_resources.start_experiment()
            result = self._executor.execute(
                _EntropyContextFactory(
                    exp_id=self._id,
                    data_writer=self._data_writer,
                    experiment_resources=self._experiment_resources,
                )
            )
            if result:
                self._data_writer.save_result(
                    self._id,
                    RawResultData(
                        "experiment_result",
                        result,
                        -1,
                        story="Final output of the experiment",
                    ),
                )
        finally:
            self._experiment_resources.end_experiment()

        self._end_time = datetime.now()

        success = True
        end_data = ExperimentEndData(self._end_time, success)
        self._data_writer.save_experiment_end_data(self._id, end_data)
        if self._executor.failed:
            raise RuntimeError("failed to execute entropy experiment")
        logger.info("Finished entropy experiment execution successfully")
        return success

    def data_reader(self) -> DataReader:
        """
        Results reader of this experiment instance
        """
        if isinstance(self._data_writer, DataReader):
            return self._data_writer
        else:
            raise Exception("database has not implemented data reader interface")

    @property
    def exp_id(self):
        return self._id


class ExperimentHandle(abc.ABC):
    """
    An handle of the experiment execution
    can be used to get information and read results
    """

    def __init__(self) -> None:
        super().__init__()

    @property
    @abc.abstractmethod
    def id(self):
        """
        experiment execution id
        """
        pass

    @property
    @abc.abstractmethod
    def results(self):
        """
        a reader for the current experiments results
        """
        pass


class ExperimentDefinition(abc.ABC):
    """
    Interface for experiment definitions.
    A definition will include the model and code for the experiment itself,
        resources and extra information.
    A definition that runs within entropy should implement all abstract methods.
    Those methods will be used inside entropy for running, logging, saving
        data and handling resources.
    """

    def __init__(
        self,
        resources: Optional[ExperimentResources],
        label: Optional[str] = None,
        story: str = None,
        user: str = "",
    ) -> None:
        super().__init__()
        self._resources: ExperimentResources = resources
        if self._resources is None:
            self._resources = ExperimentResources()
        self.label = label
        self.story = story
        self._kwargs = {}
        if user is None:
            self._user = ""
        else:
            self._user = user

    def get_experiment_resources(self) -> ExperimentResources:
        """
            experiment resources used in this definition
        :return: experiment resources used in this definition
        :rtype: ExperimentResources
        """
        return self._resources

    def _run(self, db: Optional[DataWriter] = None, **kwargs) -> _Experiment:
        """
            Create a new experiment instance using the current definition,
                and runs it
        :param db: Results db implementation
        :param kwargs: keyword parameters that can be passed to the user code
        :return: The instance of the experiment
        :rtype: _Experiment
        """
        if db is None and (not self._resources or not self._resources.get_results_db()):
            logger.warn(
                f"Results of current execution {self.label} "
                f"will be permanently lost on session close"
            )
            db = MemoryOnlyDataReaderWriter()
        elif db is None:
            db = self._resources.get_results_db()
        self._kwargs = kwargs
        experiment = _Experiment(self, db)
        experiment.run()
        return experiment

    @abc.abstractmethod
    def _get_execution_instructions(self) -> ExperimentExecutor:
        """
        The function should return a subclass of ExperimentExecutor.
        This class holds the information about executing the current type of
            experiment definition
        """
        pass

    @abc.abstractmethod
    def run(self, db: Optional[DataWriter] = None, **kwargs) -> ExperimentHandle:
        """
            Run the experiment in Entropy environment.
            Every call to this function creates a new run and returns a different handle.

        :param db: results db. if given, results will be saved in this DB. otherwise
                results will only be saved during this python session
        :param kwargs: key word arguments that will be passed to the experiment code as well.
                        user can specify here extra arguments, and request them in the
                        functions declarations.
        :return a handle of the new experiment run
        """
        pass

    @abc.abstractmethod
    def serialize(self) -> str:
        """
        Serialize the experiment definition to a string
        This string should be readable, and will be saved on every execution for debugging
            and documentation purposes.
        """
        pass



from entropylab.flame.inputs import Inputs
from entropylab.flame.workflow import Workflow

__all__ = ["ExternalScan"]


class ExternalScan(object):
    def __init__(self, workflow_node_unique_name, set_point=None):
        """Sets and maintains instrument variables

        :param set_point: (status - STREAM) node will execute only if it receives positive calibration status
        """
        self._command = "python3"
        self._bin = "externalScan.py"
        self._name = workflow_node_unique_name
        self._icon = "bootstrap/sliders.svg"
        self._inputs = _Inputs(
            set_point=set_point,
        )
        self._outputs = _Outputs(self._name)
        self._host = {}
        Workflow._register_node(self)  # register the node in the workflow context

    def host(self, **kwargs):
        """Sets additional options for execution on the host."""
        for key, value in kwargs.items():
            self._host[key] = value
        return self

    @property
    def i(self):
        """Node inputs"""
        return self._inputs

    @property
    def o(self):
        """Node outputs"""
        return self._outputs


class _Inputs(object):
    def __init__(self, set_point=None):
        self._inputs = Inputs()

        self._inputs.state(
            "set_point",
            description="node will execute only if it receives positive calibration status",
            units="status",
        )
        self._inputs.set(set_point=set_point)

    @property
    def set_point(self):
        """Input: node will execute only if it receives positive calibration status (status)"""
        return self._inputs.get("set_point")

    @set_point.setter
    def set_point(self, value):
        """Input: node will execute only if it receives positive calibration status (status)"""
        self._inputs.set(set_point=value)


class _Outputs(object):
    def __init__(self, name):
        self._name = name
        self._outputs = [
            "setpoint_reached",
        ]

    @property
    def setpoint_reached(self):
        """Output: Trigger
        :return: (bool)
        """
        return "#" + self._name + "/setpoint_reached"



import entropylab.flame.nodeio as nodeio
import time
from math import floor

# ==================== DEFINE NODE ====================

nodeio.context(
    name="BayesianEstimation",
    description="error correction algorithm",
    icon="bootstrap/speedometer.svg",
)

input = nodeio.Inputs()
input.stream("precorrected_data", units="JSON", description="cirucuit description")

output = nodeio.Outputs()
output.define(
    "correction_data",
    units="JSON",
    description="corrected data",
    retention=2,
)

nodeio.register()  # enables this to be used as part of workflow

# ==================== DRY-RUN DATA ====================

input.set(
    precorrected_data=[
        [126.05530079997635, 125.08083240358124],
        [128.8241407925009, 123.69210583423587],
    ]
)

# =============== RUN NODE STATE MACHINE ===============

v = 0

while nodeio.status.active:
    data = input.get("precorrected_data")
    print(data)
    for i in [0, 1]:
        for j in [0, 1]:
            data[i][j] = floor(data[i][j])
    print(data)

    time.sleep(0.8)
    output.set(correction_data=data)
    v += 1

# ================= =================== =================



class EntropyError(RuntimeError):
    def __init__(self, message: str, *args: object) -> None:
        super().__init__(message, *args)


class ResourceNotFound(BaseException):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



import entropylab.flame.nodeio as nodeio

# ==================== DEFINE NODE ====================

nodeio.context(
    name="ExternalScan",
    description="Sets and maintains instrument variables",
    icon="bootstrap/sliders.svg",
)

input = nodeio.Inputs()
input.stream(
    "set_point",
    units="status",
    description="node will execute only if it receives positive calibration status",
)

output = nodeio.Outputs()
output.define("setpoint_reached", units="bool", description="Trigger", retention=0)

nodeio.register()  # enables this to be used as part of workflow

# ==================== DRY-RUN DATA ====================

input.set(set_point=3.2)

# =============== RUN NODE STATE MACHINE ===============

while nodeio.status.active:
    target = input.get("set_point")

    print(f"calibrating, target = {target}")

    # time.sleep(0.01)
    output.set(setpoint_reached=True)

# ================= =================== =================



import asyncio
import enum
import sys
import time
import traceback
from copy import deepcopy
from datetime import datetime
from inspect import signature, iscoroutinefunction, getfullargspec
from itertools import count
from typing import Optional, Dict, Any, Set, Union, Callable, Coroutine, Iterable

from graphviz import Digraph

from entropylab.pipeline.api.data_reader import (
    DataReader,
    NodeResults,
    ExperimentReader,
)
from entropylab.pipeline.api.data_writer import DataWriter, NodeData
from entropylab.pipeline.api.errors import EntropyError
from entropylab.pipeline.api.execution import (
    ExperimentExecutor,
    EntropyContext,
    _EntropyContextFactory,
)
from entropylab.pipeline.api.experiment import (
    ExperimentDefinition,
    ExperimentHandle,
    _Experiment,
)
from entropylab.pipeline.api.graph import (
    GraphHelper,
    Node,
    Output,
    _NodeExecutionInfo,
    RetryBehavior,
)
from entropylab.components.lab_topology import ExperimentResources
from entropylab.logger import logger


def _handle_wait_time(wait_time, backoff, added_delay, maximum_wait_time):
    wait_time *= backoff
    wait_time += added_delay
    if maximum_wait_time is not None:
        wait_time = min(wait_time, maximum_wait_time)
    return wait_time


def _retry(
    node_label: str,
    function,
    number_of_attempts,
    wait_time,
    maximum_wait_time=None,
    backoff: float = 1,
    added_delay: float = 0,
):
    for attempt in range(number_of_attempts):
        try:
            return function()
        except BaseException as e:
            if attempt == number_of_attempts:
                raise e
            else:
                logger.warning(
                    f"node {node_label} has error, retrying #{attempt + 1}"
                    f" in {wait_time} seconds : {e}"
                )
                time.sleep(wait_time)
                wait_time = _handle_wait_time(
                    wait_time, backoff, added_delay, maximum_wait_time
                )


def pynode(
    label: str,
    input_vars: Dict[str, Output] = None,
    output_vars: Set[str] = None,
    must_run_after: Set[Node] = None,
):
    """
        decorator for running using the given python function as a PyNode
    :param label: node label
    :param input_vars: dictionary of node inputs, keys are input names.
                        the input values should be defined as the following:
                            >>>input_vars={"a": node.outputs["x"]}
    :param output_vars: a set of outputs name. The given python function should
                        return a dictionary, which it's keys are the same as output vars
    :param must_run_after: A set of nodes. If those nodes are in the same graph,
                            current node will run after they finish execution.
    :return: node instance
    """

    def decorate(fn):
        return PyNode(label, fn, input_vars, output_vars, must_run_after)

    return decorate


class PyNode(Node):
    """
    Node that gets a python function or coroutine and wraps it with an entropy graph node.
    Node is defined by a python function, input and outputs.
    Entropy will call the function, filling the function parameters:
        1. If function parameter has the same name as node input,
           Entropy will pass the input value
        2. If function parameter is of type EntropyContext, Entropy will pass current context
        3. If function parameter has the same name as given experiment run kwargs,
           Entropy will pass the kwarg value
        4. if function parameter is of type *args,
           Entropy will pass all inputs that were not specified by name
        5. If function parameter name is "is_last",
           Entropy will pass True if this is the last node in the graph

    """

    def __init__(
        self,
        label: str = None,
        program: Union[Callable, Coroutine] = None,
        input_vars: Dict[str, Output] = None,
        output_vars: Set[str] = None,
        must_run_after: Set[Node] = None,
        save_results: bool = True,
        retry_on_error: RetryBehavior = None,
    ):
        """
            Node that gets a python function or coroutine and wraps
            it with an entropy graph node.
            Node is defined by a python function, input and outputs.
            Entropy will call the function, filling the function parameters:
                1. If function parameter has the same name as node input,
                   Entropy will pass the input value
                2. If function parameter is of type EntropyContext,
                   Entropy will pass current context
                3. If function parameter has the same name as given
                   experiment run kwargs, Entropy will pass the kwarg value
                4. if function parameter is of type *args,
                   Entropy will pass all inputs that were not specified by name
                5. If function parameter name is "is_last",
                   Entropy will pass True if this is the last node in the graph
        :param label:  node label
        :param program: the node program in a python function or coroutine
        :param input_vars: dictionary of node inputs, keys are input names.
                        the input values should be defined as the following:
                            >>>input_vars={"a": node.outputs["x"]}
        :param output_vars: a set of outputs name. The given python function should
                        return a dictionary, which it's keys are the same as output vars
        :param must_run_after: A set of nodes. If those nodes are in the same graph,
                            current node will run after they finish execution.
        """
        super().__init__(
            label, input_vars, output_vars, must_run_after, save_results, retry_on_error
        )
        self._program = program

    async def _execute_async(
        self,
        input_values: Dict[str, Any],
        context: EntropyContext,
        is_last,
        **kwargs,
    ) -> Dict[str, Any]:
        args_parameters, keyword_function_parameters = self._prepare_for_execution(
            context, is_last, kwargs, input_values
        )

        try:
            if iscoroutinefunction(self._program):
                results = await self._program(
                    *args_parameters, **keyword_function_parameters
                )
            else:
                results = self._program(*args_parameters, **keyword_function_parameters)

            return self._handle_results(results)
        except BaseException as e:
            raise e

    def _execute(
        self,
        input_values: Dict[str, Any],
        context: EntropyContext,
        is_last,
        **kwargs,
    ) -> Dict[str, Any]:
        args_parameters, keyword_function_parameters = self._prepare_for_execution(
            context, is_last, kwargs, input_values
        )

        try:
            if iscoroutinefunction(self._program):
                results = asyncio.run(
                    self._program(*args_parameters, **keyword_function_parameters)
                )
            else:
                results = self._program(*args_parameters, **keyword_function_parameters)

            return self._handle_results(results)
        except BaseException as e:
            raise e

    def _handle_results(self, results):
        outputs = {}
        if isinstance(results, Dict):
            for var in self._output_vars:
                try:
                    outputs[var] = results[var]
                except KeyError:
                    logger.error(
                        f"WARNING Could not fetch variable '{var}' "
                        f"from the results of node <{self.label}>"
                    )
                    pass
        else:
            if results:
                raise EntropyError(
                    f"node {self.label} result should be a "
                    f"dictionary but is {type(results)}"
                )
        return outputs

    def _prepare_for_execution(
        self, context, is_last, kwargs, input_values: Dict[str, Any]
    ):
        sig = signature(self._program)
        keyword_function_parameters = {}
        for param in sig.parameters:
            if sig.parameters[param].annotation is EntropyContext:
                keyword_function_parameters[param] = context
        if "is_last" in sig.parameters:
            keyword_function_parameters["is_last"] = is_last
        (
            args,
            varargs,
            varkw,
            defaults,
            kwonlyargs,
            kwonlydefaults,
            annotations,
        ) = getfullargspec(self._program)
        for arg in args + kwonlyargs:
            if (
                arg not in input_values
                and arg not in keyword_function_parameters
                and arg not in kwargs
                and (
                    (defaults and arg not in defaults)
                    or (kwonlydefaults and arg not in kwonlydefaults)
                )
            ):
                logger.error(f"Error in node {self.label} - {arg} is not in parameters")
                raise KeyError(arg)

            if arg in input_values:
                keyword_function_parameters[arg] = input_values[arg]
            elif arg in kwargs:
                keyword_function_parameters[arg] = kwargs[arg]
        args_parameters = []
        if varargs is not None:
            for item in input_values:
                if item not in keyword_function_parameters:
                    args_parameters.append(input_values[item])
        return args_parameters, keyword_function_parameters


def _create_actual_graph(nodes: Set[Node], key_nodes: Set[Node]):
    nodes_copy: Dict[Node, Node] = {node: deepcopy(node) for node in nodes}
    for node in nodes_copy:
        for input_var in nodes_copy[node]._input_vars:
            output = node._input_vars[input_var]
            nodes_copy[node]._input_vars[input_var] = nodes_copy[output.node].outputs[
                output.name
            ]
        nodes_copy[node]._must_run_after = {nodes_copy[m] for m in node._must_run_after}
    return {_NodeExecutionInfo(nodes_copy[node], node in key_nodes) for node in nodes}


class SubGraphNode(Node):
    """
    Node that holds a complete graph and runs as a single node within
    another graph.
    """

    def __init__(
        self,
        graph: Union[GraphHelper, Node, Set[Node]],
        label: str = None,
        input_vars: Dict[str, Output] = None,
        output_vars: Set[str] = None,
        must_run_after: Set[Node] = None,
        key_nodes: Optional[Set[Node]] = None,
        save_results: bool = True,
        retry_on_error: RetryBehavior = None,
    ):
        """

        :param graph: the graph model
        :param label: node label
        :param input_vars: dictionary of node inputs, keys are input names.
                        the input values should be defined as the following:
                            >>>input_vars={"a": node.outputs["x"]}
        :param output_vars: a set of outputs name. The given python function should
                        return a dictionary, which it's keys are the same as output vars
        :param must_run_after: A set of nodes. If those nodes are in the same graph,
                            current node will run after they finish execution.
        """
        super().__init__(
            label, input_vars, output_vars, must_run_after, save_results, retry_on_error
        )
        self._key_nodes = key_nodes
        if self._key_nodes is None:
            self._key_nodes = set()

        if isinstance(graph, GraphHelper):
            self._graph: Set[_NodeExecutionInfo] = _create_actual_graph(
                graph.nodes, self._key_nodes
            )
        elif isinstance(graph, Node):
            self._graph: Set[_NodeExecutionInfo] = _create_actual_graph(
                {graph}, self._key_nodes
            )
        elif isinstance(graph, Set):
            self._graph: Set[_NodeExecutionInfo] = _create_actual_graph(
                graph, self._key_nodes
            )
        else:
            raise Exception(
                "graph parameter type is not supported, please pass a Node or set of nodes"
            )

    async def _execute_async(
        self,
        input_values: Dict[str, Any],
        context: EntropyContext,
        is_last,
        **kwargs,
    ) -> Dict[str, Any]:
        executors = {node.node: _NodeExecutor(node) for node in self._graph}
        return await _AsyncGraphExecutor(
            self._graph, executors, **kwargs
        ).execute_async(context._context_factory)

    def _execute(
        self,
        input_values: Dict[str, Any],
        context: EntropyContext,
        is_last,
        **kwargs,
    ) -> Dict[str, Any]:
        executors = {node.node: _NodeExecutor(node) for node in self._graph}
        return _GraphExecutor(self._graph, executors, **kwargs).execute(
            context._context_factory
        )


class _NodeExecutor:
    def __init__(self, node_execution_info: _NodeExecutionInfo) -> None:
        super().__init__()
        self._node: Node = node_execution_info.node
        self._start_time: Optional[datetime] = None
        self._end_time: Optional[datetime] = None
        self.result: Dict[str, Any] = {}
        self.to_run = True
        self._is_key_node = node_execution_info.is_key_node

    def run(
        self,
        input_values: Dict[str, Any],
        context_factory: _EntropyContextFactory,
        is_last: int,
        **kwargs,
    ) -> Dict[str, Any]:
        if self.to_run:
            context = context_factory.create()
            self._prepare_for_run(context)
            retry_behavior = self._node._retry_on_error_function()
            if retry_behavior is not None:
                self.result = _retry(
                    self._node.label,
                    lambda: self._node._execute(
                        input_values,
                        context,
                        is_last,
                        **kwargs,
                    ),
                    number_of_attempts=retry_behavior.number_of_attempts,
                    wait_time=retry_behavior.wait_time,
                    backoff=retry_behavior.backoff,
                    added_delay=retry_behavior.added_delay,
                    maximum_wait_time=retry_behavior.max_wait_time,
                )
            else:
                self.result = self._node._execute(
                    input_values,
                    context,
                    is_last,
                    **kwargs,
                )
            return self._handle_result(context)

    async def run_async(
        self,
        input_values: Dict[str, Any],
        context_factory: _EntropyContextFactory,
        is_last: int,
        **kwargs,
    ) -> Dict[str, Any]:
        if self.to_run:
            context = context_factory.create()
            self._prepare_for_run(context)
            retry_behavior = self._node._retry_on_error_function()
            if retry_behavior is not None:
                self.result = await _retry(
                    self._node.label,
                    lambda: self._node._execute_async(
                        input_values,
                        context,
                        is_last,
                        **kwargs,
                    ),
                    number_of_attempts=retry_behavior.number_of_attempts,
                    wait_time=retry_behavior.wait_time,
                    backoff=retry_behavior.backoff,
                    added_delay=retry_behavior.added_delay,
                    maximum_wait_time=retry_behavior.max_wait_time,
                )
            else:
                self.result = await self._node._execute_async(
                    input_values,
                    context,
                    is_last,
                    **kwargs,
                )
            return self._handle_result(context)

    def _handle_result(self, context):
        if self._node._should_save_results():
            # logger fetching results
            for output_id in self.result:
                output = self.result[output_id]
                context.add_result(label=f"{output_id}", data=output)

        self._end_time = datetime.now()
        logger.debug(
            f"Done running node <{self._node.__class__.__name__}> {self._node.label}"
        )
        return self.result

    def _prepare_for_run(self, context: EntropyContext):
        logger.info(
            f"Running node <{self._node.__class__.__name__}> {self._node.label}"
        )
        self._start_time = datetime.now()
        logger.debug(
            f"Saving metadata before running node "
            f"<{self._node.__class__.__name__}> {self._node.label} id={context._get_stage_id()}"
        )
        context._data_writer.save_node(
            context._exp_id,
            NodeData(
                context._get_stage_id(),
                self._start_time,
                self._node.label,
                self._is_key_node,
            ),
        )


class _AsyncGraphExecutor(ExperimentExecutor):
    def __init__(
        self,
        nodes_execution_info: Set[_NodeExecutionInfo],
        nodes: Dict[Node, _NodeExecutor],
        **kwargs,
    ) -> None:
        super().__init__()
        self._graph: GraphHelper = GraphHelper(nodes_execution_info)
        self._node_kwargs = kwargs
        self._tasks: Dict[Node, Any] = dict()
        self._results: Dict = dict()
        self._stopped = False
        self._node_id_iter = count(start=0, step=1)
        self._executors: Dict[Node, _NodeExecutor] = nodes

    def execute(self, context: _EntropyContextFactory) -> Any:
        async_result = asyncio.run(self.execute_async(context))
        return async_result

    @property
    def failed(self) -> bool:
        return self._stopped

    async def execute_async(self, context_factory: _EntropyContextFactory):
        # traverse the graph and run the nodes
        end_nodes = self._graph.leaves

        chains = []
        for node in end_nodes:
            chains.append(self._run_node_and_ancestors(node, context_factory, 0))

        result = await asyncio.gather(*chains)
        result = [x for x in result if x is not None]
        combined_result = {}

        if result:
            combined_result = {k: v for d in result for k, v in d.items()}
        return combined_result

    async def _run_node_and_ancestors(
        self, node: Node, context_factory: _EntropyContextFactory, is_last: int
    ):
        tasks = []
        for input_name in node.get_parents():
            if input_name not in self._tasks:
                task = self._run_node_and_ancestors(
                    input_name, context_factory, is_last + 1
                )
                tasks.append(task)
                self._tasks[input_name] = task
            else:
                tasks.append(self._tasks[input_name])
        results = []
        if len(tasks) > 0:
            await asyncio.wait(tasks)
            if self._stopped:
                return None
            results = {}
            inputs_by_name = node.get_inputs_by_name()
            for input_name in inputs_by_name:
                parent_node = inputs_by_name[input_name].node
                parent_output_name = inputs_by_name[input_name].name
                if (
                    parent_node not in self._executors
                    or parent_output_name not in self._executors[parent_node].result
                ):
                    raise EntropyError(
                        f"node {node.label} input is missing: {parent_output_name}"
                    )
                results[input_name] = self._executors[parent_node].result[
                    parent_output_name
                ]

        node_executor = self._executors[node]
        try:
            return await node_executor.run_async(
                results,
                context_factory,
                node in self._graph.leaves,
                **self._node_kwargs,
            )
        except BaseException as e:
            self._stopped = True
            trace = "\n".join(traceback.format_exception(*sys.exc_info()))
            logger.error(
                f"Stopping GraphHelper, Error in node {node.label} "
                f"of type {e.__class__.__qualname__}. message: {e}\ntrace:\n{trace}"
            )
            return


class _GraphExecutor(ExperimentExecutor):
    def __init__(
        self,
        nodes_execution_info: Set[_NodeExecutionInfo],
        nodes: Dict[Node, _NodeExecutor],
        **kwargs,
    ) -> None:
        super().__init__()
        self._graph: GraphHelper = GraphHelper(nodes_execution_info)
        self._node_kwargs = kwargs
        self._tasks: Dict[Node, Any] = dict()
        self._results: Dict = dict()
        self._stopped = False
        self._executors: Dict[Node, _NodeExecutor] = nodes

    def execute(self, context_factory: _EntropyContextFactory) -> Any:
        sorted_nodes = self._graph.nodes_in_topological_order()

        for node in sorted_nodes:
            results = {}
            inputs_by_name = node.get_inputs_by_name()
            for input_name in inputs_by_name:
                parent_node = inputs_by_name[input_name].node
                parent_output_name = inputs_by_name[input_name].name
                if (
                    parent_node not in self._executors
                    or parent_output_name not in self._executors[parent_node].result
                ):
                    raise EntropyError(
                        f"node {node.label} input is missing: {parent_output_name}"
                    )
                results[input_name] = self._executors[parent_node].result[
                    parent_output_name
                ]
            node_executor = self._executors[node]
            try:
                node_executor.run(
                    results,
                    context_factory,
                    node in self._graph.leaves,
                    **self._node_kwargs,
                )
            except BaseException as e:
                self._stopped = True
                trace = "\n".join(traceback.format_exception(*sys.exc_info()))
                logger.error(
                    f"Stopping GraphHelper, Error in node {node.label} of type "
                    f"{e.__class__.__qualname__}. message: {e}\ntrace:\n{trace}"
                )
                return
        combined_result = {}
        for node in self._graph.leaves:
            result = self._executors[node].result
            if result:
                for key in result:
                    combined_result[key] = result[key]

        return combined_result

    @property
    def failed(self) -> bool:
        return self._stopped


class GraphReader(ExperimentReader):
    """
    Reads results and data from a single graph experiment
    """

    def __init__(self, experiment_id: int, db: DataReader) -> None:
        """
            Reads results and data from a single graph experiment
        :param experiment_id: the id of experiment
        :param db: results database that implemented the DataReader abstract class
        """
        super().__init__(experiment_id, db)

    def get_results_from_node(
        self, node_label: str, result_label: Optional[str] = None
    ) -> Iterable[NodeResults]:
        """
            returns an iterable of all results from a node with the given label
             and result with the given result_label
        :param node_label: label of node to get results from
        :param result_label: label of result records
        """
        return self._data_reader.get_results_from_node(
            node_label, self._experiment_id, result_label
        )


class GraphExperimentHandle(ExperimentHandle):
    """
    An handle of the graph experiment execution
    can be used to get information and read results
    """

    def __init__(self, experiment: _Experiment, graph: GraphHelper) -> None:
        super().__init__()
        self._experiment = experiment
        self._graph = graph

    @property
    def id(self):
        return self._experiment.exp_id

    @property
    def results(self) -> GraphReader:
        """
        returns a reader for reading results from a graph experiments
        """
        return GraphReader(self.id, self._experiment.data_reader())

    def dot_graph(self):
        return self._graph.export_dot_graph()


class GraphExecutionType(enum.Enum):
    Sync = 1
    Async = 2


class Graph(ExperimentDefinition):
    """
    Experiment defined by a graph model and runs within entropy.
    Information, results and metadata will be saved during every execution.
    Nodes within the graph will be executed in a topological order.
    Every node will be declared with a set of inputs and outputs, that
    will transfer for a node to it's children.
    """

    def __init__(
        self,
        resources: Optional[ExperimentResources],
        graph: Union[Node, Set[Node], GraphHelper],
        label: Optional[str] = None,
        story: str = None,
        key_nodes: Optional[Set[Node]] = None,
        execution_type: GraphExecutionType = GraphExecutionType.Sync,
        user: str = "",
    ) -> None:
        """
            Experiment defined by a graph model and runs within entropy.
        :param resources: shared lab resources or temporary resources that
                            are used in the experiment.
        :param graph: the experiment model, can be a graph or a single node.
        :param label: experiment label
        :param story: a description of the experiment, which will create an experiment story
                         with all other parts of the experiment
        :param key_nodes: a set of graph key nodes. those nodes will be marked as graph result.
        :param execution_type: specifty whether to run the graph in a sync mode, single node
                        on a given time, or asynchronously - which will run node in parallel
                        according to their dependency and implementation (using async.io)
        """
        super().__init__(resources, label, story, user)
        self._key_nodes = key_nodes
        if self._key_nodes is None:
            self._key_nodes = set()

        if isinstance(graph, GraphHelper):
            self._original_nodes: Set[Node] = graph.nodes
        elif isinstance(graph, Node):
            self._original_nodes: Set[Node] = {graph}
        elif isinstance(graph, Set):
            self._original_nodes: Set[Node] = graph
        else:
            raise Exception(
                "graph parameter type is not supported, please pass a Node or set of nodes"
            )

        all_ancestors = set(
            [anc for node in self._original_nodes for anc in node.ancestors()]
        )
        if all_ancestors != self._original_nodes:
            raise Exception(
                f"nodes has inputs that are not part of this graph: "
                f"{[node.label for node in all_ancestors.difference(self._original_nodes)]}"
            )

        self._actual_graph: Set[_NodeExecutionInfo] = _create_actual_graph(
            self._original_nodes, self._key_nodes
        )
        self._to_node: Optional[Node] = None
        self._execution_type: GraphExecutionType = execution_type

    def _get_execution_instructions(self) -> ExperimentExecutor:
        executors = {node.node: _NodeExecutor(node) for node in self._actual_graph}

        if self._execution_type == GraphExecutionType.Sync:
            return _GraphExecutor(self._actual_graph, executors, **self._kwargs)
        elif self._execution_type == GraphExecutionType.Async:
            return _AsyncGraphExecutor(self._actual_graph, executors, **self._kwargs)
        else:
            raise Exception(f"Execution type {self._execution_type} is not supported")

    def serialize(self) -> str:
        """
        dot graph representing the experiment
        """
        return str(GraphHelper(self._actual_graph).export_dot_graph())

    def run(self, db: Optional[DataWriter] = None, **kwargs) -> GraphExperimentHandle:
        experiment = self._run(db, **kwargs)
        return GraphExperimentHandle(experiment, GraphHelper(self._actual_graph))

    def run_to_node(
        self,
        node: Node,
        db: Optional[DataWriter] = None,
        label: Optional[str] = None,
        **kwargs,
    ) -> GraphExperimentHandle:
        """
            Run the experiment in Entropy environment, only with the given node and
            its ancestors.
            Every call to this function creates a new run and returns a different handle.

        :param node: the node object you want to run to.
        :param db: results db. if given, results will be saved in this DB. otherwise
                results will only be saved during this python session
        :param label: label for the current execution
        :param kwargs: key word arguments that will be passed to the experiment code as well.
                        user can specify here extra arguments, and request them in the
                        functions declarations.
        :return:a handle of the new graph experiment run
        """
        if node not in self._original_nodes:
            raise KeyError("Node is not in graph")
        logger.info(f"Running node {node.label} and dependencies")
        nodes = self._calculate_ancestors(node)
        full_graph = self._actual_graph
        self._actual_graph = _create_actual_graph(nodes, self._key_nodes)
        old_label = self.label
        if label:
            self.label = label
        try:
            return self.run(db, **kwargs)
        finally:
            self.label = old_label
            self._actual_graph = full_graph

    def _calculate_ancestors(self, node):
        ancestors: Set = set()
        for parent in node.ancestors():
            if parent in self._original_nodes:
                ancestors.add(parent)
        return ancestors

    def dot_graph(self) -> Digraph:
        return GraphHelper(self._actual_graph).export_dot_graph()
